import autocti as ac

from test_autocti.simulators import resolution_util


def simulate__ci_uniform__parallel_x1(resolution, normalizations):

    ci_data_type = "ci_uniform"
    ci_data_model = "parallel_x1"

    parallel_traps = [ac.TrapInstantCapture(density=1.0, release_timescale=3.0)]
    parallel_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    clocker = ac.Clocker(iterations=1, parallel_express=2)

    for normalization in normalizations:

        ci_regions = resolution_util.ci_regions_from_resolution(resolution=resolution)
        pattern = ac.ci.CIPatternUniform(
            normalization=normalization, regions=ci_regions
        )

        resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
            ci_data_type=ci_data_type,
            ci_data_model=ci_data_model,
            resolution=resolution,
            clocker=clocker,
            pattern=pattern,
            parallel_traps=parallel_traps,
            parallel_ccd=parallel_ccd,
        )


def simulate__ci_uniform__parallel_x3(resolution, normalizations):

    ci_data_type = "ci_uniform"
    ci_data_model = "parallel_x3"

    parallel_traps = [
        ac.TrapInstantCapture(density=0.5, release_timescale=2.0),
        ac.TrapInstantCapture(density=1.5, release_timescale=5.0),
        ac.TrapInstantCapture(density=2.5, release_timescale=20.0),
    ]

    parallel_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    clocker = ac.Clocker(iterations=1, parallel_express=2)

    for normalization in normalizations:
        ci_regions = resolution_util.ci_regions_from_resolution(resolution=resolution)
        pattern = ac.ci.CIPatternUniform(
            normalization=normalization, regions=ci_regions
        )

        resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
            ci_data_type=ci_data_type,
            ci_data_model=ci_data_model,
            resolution=resolution,
            pattern=pattern,
            parallel_traps=parallel_traps,
            parallel_ccd=parallel_ccd,
            clocker=clocker,
        )


def simulate__ci_uniform__serial_x1(resolution, normalizations):

    ci_data_type = "ci_uniform"
    ci_data_model = "serial_x1"

    serial_traps = [ac.TrapInstantCapture(density=1.0, release_timescale=3.0)]
    serial_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    clocker = ac.Clocker(iterations=1, serial_express=2)

    for normalization in normalizations:

        ci_regions = resolution_util.ci_regions_from_resolution(resolution=resolution)
        pattern = ac.ci.CIPatternUniform(
            normalization=normalization, regions=ci_regions
        )

        resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
            ci_data_type=ci_data_type,
            ci_data_model=ci_data_model,
            resolution=resolution,
            pattern=pattern,
            serial_traps=serial_traps,
            serial_ccd=serial_ccd,
            clocker=clocker,
        )


def simulate__ci_uniform__serial_x3(resolution, normalizations):

    ci_data_type = "ci_uniform"
    ci_data_model = "serial_x3"

    serial_traps = [
        ac.TrapInstantCapture(density=0.5, release_timescale=2.0),
        ac.TrapInstantCapture(density=1.5, release_timescale=5.0),
        ac.TrapInstantCapture(density=2.5, release_timescale=20.0),
    ]
    serial_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    clocker = ac.Clocker(iterations=1, serial_express=2)

    for normalization in normalizations:
        ci_regions = resolution_util.ci_regions_from_resolution(resolution=resolution)
        pattern = ac.ci.CIPatternUniform(
            normalization=normalization, regions=ci_regions
        )

        resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
            ci_data_type=ci_data_type,
            ci_data_model=ci_data_model,
            resolution=resolution,
            pattern=pattern,
            serial_traps=serial_traps,
            serial_ccd=serial_ccd,
            clocker=clocker,
        )


def simulate__ci_uniform__parallel_x1__serial_x1(resolution, normalizations):

    ci_data_type = "ci_uniform"
    ci_data_model = "parallel_x1__serial_x1"

    parallel_traps = [ac.TrapInstantCapture(density=1.0, release_timescale=3.0)]
    parallel_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    serial_traps = [ac.TrapInstantCapture(density=1.0, release_timescale=3.0)]
    serial_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    clocker = ac.Clocker(iterations=1, parallel_express=2, serial_express=2)

    for normalization in normalizations:

        ci_regions = resolution_util.ci_regions_from_resolution(resolution=resolution)
        pattern = ac.ci.CIPatternUniform(
            normalization=normalization, regions=ci_regions
        )

        resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
            ci_data_type=ci_data_type,
            ci_data_model=ci_data_model,
            resolution=resolution,
            pattern=pattern,
            parallel_traps=parallel_traps,
            parallel_ccd=parallel_ccd,
            serial_traps=serial_traps,
            serial_ccd=serial_ccd,
            clocker=clocker,
        )


def simulate__ci_uniform__parallel_x3__serial_x3(resolution, normalizations):

    ci_data_type = "ci_uniform"
    ci_data_model = "parallel_x3__serial_x3"

    parallel_traps = [
        ac.TrapInstantCapture(density=0.5, release_timescale=2.0),
        ac.TrapInstantCapture(density=1.5, release_timescale=5.0),
        ac.TrapInstantCapture(density=2.5, release_timescale=20.0),
    ]

    parallel_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    serial_traps = [
        ac.TrapInstantCapture(density=0.5, release_timescale=2.0),
        ac.TrapInstantCapture(density=1.5, release_timescale=5.0),
        ac.TrapInstantCapture(density=2.5, release_timescale=20.0),
    ]
    serial_ccd = ac.CCD(
        well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
    )

    clocker = ac.Clocker(iterations=1, parallel_express=2, serial_express=2)

    for normalization in normalizations:
        ci_regions = resolution_util.ci_regions_from_resolution(resolution=resolution)
        pattern = ac.ci.CIPatternUniform(
            normalization=normalization, regions=ci_regions
        )

        resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
            ci_data_type=ci_data_type,
            ci_data_model=ci_data_model,
            resolution=resolution,
            pattern=pattern,
            parallel_traps=parallel_traps,
            parallel_ccd=parallel_ccd,
            serial_traps=serial_traps,
            serial_ccd=serial_ccd,
            clocker=clocker,
        )


# def simulate__ci_uniform_cosmic_rays__parallel_x1(resolution, normalizations):
#
#     ci_data_type = "ci_uniform_cosmic_rays"
#     ci_data_model = "parallel_x1"
#
#     parallel_traps = ac.TrapInstantCapture(density=1.0, release_timescale=3.0)
#     parallel_ccd = ac.CCD(
#         well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
#     )
#
#     clocker = ac.Clocker(iterations=1, parallel_express=2)
#
#     shape = resolution_util.shape_2d_from_resolution(
#         resolution=resolution
#     )
#
#     for normalization in normalizations:
#
#         ci_regions = resolution_util.ci_regions_from_resolution(
#             resolution=resolution
#         )
#         pattern = ac.ci.CIPatternUniform(
#             normalization=normalization, regions=ci_regions
#         )
#
#         cosmic_ray_map = cosmic_ray_map_from_shape_and_well_depth(
#             shape=shape, well_depth=parallel_ccd.well_notch_depth
#         )
#
#         return resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
#             ci_data_type=ci_data_type,
#             ci_data_model=ci_data_model,
#             resolution=resolution,
#             pattern=pattern,
#             parallel_traps=parallel_traps,
#             parallel_ccd=parallel_ccd,
#             clocker=clocker,
#             cosmic_ray_map=cosmic_ray_map,
#         )
#
#
# def simulate__ci_uniform_cosmic_rays__serial_x1(resolution, normalizations):
#
#     ci_data_type = "ci_uniform_cosmic_rays"
#     ci_data_model = "serial_x1"
#
#     serial_traps = ac.TrapInstantCapture(density=1.0, release_timescale=3.0)
#     serial_ccd = ac.CCD(
#         well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
#     )
#
#     clocker = ac.Clocker(iterations=1, serial_express=2)
#
#     shape = resolution_util.shape_2d_from_resolution(
#         resolution=resolution
#     )
#
#     for normalization in normalizations:
#
#         ci_regions = resolution_util.ci_regions_from_resolution(
#             resolution=resolution
#         )
#         pattern = ac.ci.CIPatternUniform(
#             normalization=normalization, regions=ci_regions
#         )
#
#         cosmic_ray_map = cosmic_ray_map_from_shape_and_well_depth(
#             shape=shape, well_depth=serial_ccd.well_notch_depth
#         )
#
#         resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
#             ci_data_type=ci_data_type,
#             ci_data_model=ci_data_model,
#             resolution=resolution,
#             pattern=pattern,
#             serial_traps=serial_traps,
#             serial_ccd=serial_ccd,
#             clocker=clocker,
#             cosmic_ray_map=cosmic_ray_map,
#         )
#
#
# def simulate__ci_uniform_cosmic_rays__parallel_x1__serial_x1(
#     resolution, normalizations
# ):
#
#     ci_data_type = "ci_uniform_cosmic_rays"
#     ci_data_model = "parallel_x1__serial_x1"
#
#     parallel_traps = ac.TrapInstantCapture(density=1.0, release_timescale=3.0)
#     parallel_ccd = ac.CCD(
#         well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
#     )
#
#     serial_traps = ac.TrapInstantCapture(density=1.0, release_timescale=3.0)
#     serial_ccd = ac.CCD(
#         well_fill_power=0.8, full_well_depth=8.47e4, well_notch_depth=1e-7
#     )
#
#     clocker = ac.Clocker(iterations=1, parallel_express=1, serial_express=2)
#
#     shape = resolution_util.shape_2d_from_resolution(
#         resolution=resolution
#     )
#
#     for normalization in normalizations:
#
#         ci_regions = resolution_util.ci_regions_from_resolution(
#             resolution=resolution
#         )
#         pattern = ac.ci.CIPatternUniform(
#             normalization=normalization, regions=ci_regions
#         )
#
#         cosmic_ray_map = cosmic_ray_map_from_shape_and_well_depth(
#             shape=shape, well_depth=parallel_ccd.well_notch_depth
#         )
#
#         resolution_util.simulate_ci_data_from_ci_normalization_region_and_cti_model(
#             ci_data_type=ci_data_type,
#             ci_data_model=ci_data_model,
#             resolution=resolution,
#             pattern=pattern,
#             parallel_traps=parallel_traps,
#             parallel_ccd=parallel_ccd,
#             serial_traps=serial_traps,
#             serial_ccd=serial_ccd,
#             clocker=clocker,
#             cosmic_ray_map=cosmic_ray_map,
#         )


# def cosmic_ray_map_from_shape_and_well_depth(shape, well_depth):
#     # We use the LA Cosmic algorithm to simulate and add cosmic rays to our ci pre cti image. This routine randomly
#     # generates cosmimc rays based on realistic cosmic ray rates expected. These cosmic rays will then be added to our
#     # ci pre-cti image in the simulaate function below, and subject to CTI according to the CTI model.
#     cosmic_path = af.util.create_path(
#         path=workspace_path, folder_names=["scripts", "cosmic_rays"]
#     )
#
#     cosmic_ray_maker = cosmics.CosmicRays(
#         shape=shape,
#         cr_fluxscaling=1.0,
#         cr_length_file=cosmic_path + "crlength_v2.fits",
#         cr_distance_file=cosmic_path + "crdist.fits",
#         log=logger,
#     )
#     cosmic_ray_maker.set_ifiles()
#     return cosmic_ray_maker.drawEventsToCoveringFactor(limit=well_depth)



import numpy as np

from autocti.charge_injection import ci_imaging
from autoarray.fit import fit


class CIFitImaging(fit.FitImaging):
    def __init__(
        self,
        masked_ci_imaging: ci_imaging.MaskedCIImaging,
        ci_post_cti,
        hyper_noise_scalars=None,
    ):
        """Fit a charge injection ci_data-set with a model cti image, also scalng the noises within a Bayesian
        framework.

        Parameters
        -----------
        masked_ci_imaging
            The charge injection image that is fitted.
        cti_params : arctic_params.ArcticParams
            The cti model parameters which describe how CTI during clocking.
        clocker : arctic_settings.ArcticSettings
            The settings that control how arctic models CTI.
        hyper_noise_scalars :
            The ci_hyper-parameter(s) which the noise_scaling_maps_list is multiplied by to scale the noise-map.
        """

        self.ci_masked_data = masked_ci_imaging
        self.hyper_noise_scalars = hyper_noise_scalars

        if hyper_noise_scalars is not None and len(hyper_noise_scalars) > 0:
            self.noise_scaling_maps = masked_ci_imaging.noise_scaling_maps

            noise_map = hyper_noise_map_from_noise_map_and_noise_scalings(
                hyper_noise_scalars=hyper_noise_scalars,
                noise_scaling_maps=masked_ci_imaging.noise_scaling_maps,
                noise_map=masked_ci_imaging.noise_map,
            )

            masked_ci_imaging = masked_ci_imaging.modify_noise_map(noise_map=noise_map)

        super().__init__(masked_imaging=masked_ci_imaging, model_image=ci_post_cti)

    @property
    def masked_ci_imaging(self):
        return self.ci_masked_data

    @property
    def ci_post_cti(self):
        return self.model_data

    @property
    def ci_pre_cti(self):
        return self.ci_masked_data.ci_pre_cti

    @property
    def chi_squared_map_of_ci_regions(self):
        return self.chi_squared_map.ci_regions_frame

    @property
    def chi_squared_map_of_parallel_trails(self):
        return self.chi_squared_map.parallel_non_ci_regions_frame

    @property
    def chi_squared_map_of_serial_trails(self):
        return self.chi_squared_map.serial_trails_frame

    @property
    def chi_squared_map_of_serial_overscan_no_trails(self):
        return self.chi_squared_map.serial_overscan_no_trails_frame


def hyper_noise_map_from_noise_map_and_noise_scalings(
    hyper_noise_scalars, noise_scaling_maps, noise_map
):
    """For a noise-map, use the model hyper noise and noise-scaling maps to compute a scaled noise-map.

    Parameters
    -----------
    hyper_noise_scalars
    noise_scaling_maps
    noise_map : imaging.NoiseMap or ndarray
        An arrays describing the RMS standard deviation error in each pixel, preferably in units of electrons per
        second.
    """
    hyper_noise_maps = list(
        map(
            lambda hyper_noise_scalar, noise_scaling_map: hyper_noise_scalar.scaled_noise_map_from_noise_scaling(
                noise_scaling_map
            ),
            hyper_noise_scalars,
            noise_scaling_maps,
        )
    )

    return np.add(noise_map, sum(hyper_noise_maps))



import os

import pytest

from autocti import plot as aplt

directory = os.path.dirname(os.path.realpath(__file__))


@pytest.fixture(name="plot_path")
def make_ci_imaging_plotter_setup():
    return "{}/files/plots/ci_imaging/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


def test__individual_attributes_are_output(ci_imaging_7x7, plot_path, plot_patch):

    aplt.CIImaging.image(
        ci_imaging=ci_imaging_7x7,
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/image.png" in plot_patch.paths

    aplt.CIImaging.noise_map(
        ci_imaging=ci_imaging_7x7,
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/noise_map.png" in plot_patch.paths

    aplt.CIImaging.ci_pre_cti(
        ci_imaging=ci_imaging_7x7,
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/ci_pre_cti.png" in plot_patch.paths

    aplt.CIImaging.signal_to_noise_map(
        ci_imaging=ci_imaging_7x7,
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/signal_to_noise_map.png" in plot_patch.paths

    ci_imaging_7x7.cosmic_ray_map[0, 0] = 1.0

    aplt.CIImaging.cosmic_ray_map(
        ci_imaging=ci_imaging_7x7,
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )
    assert f"{plot_path}/cosmic_ray_map.png" in plot_patch.paths


def test__individual_lines_are_output(ci_imaging_7x7, plot_path, plot_patch):

    aplt.CIImaging.image_line(
        ci_imaging=ci_imaging_7x7,
        line_region="parallel_front_edge",
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/image_line.png" in plot_patch.paths

    aplt.CIImaging.noise_map_line(
        ci_imaging=ci_imaging_7x7,
        line_region="parallel_front_edge",
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/noise_map_line.png" in plot_patch.paths

    aplt.CIImaging.ci_pre_cti_line(
        ci_imaging=ci_imaging_7x7,
        line_region="parallel_front_edge",
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/ci_pre_cti_line.png" in plot_patch.paths

    aplt.CIImaging.signal_to_noise_map_line(
        ci_imaging=ci_imaging_7x7,
        line_region="parallel_front_edge",
        include=aplt.Include(),
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/signal_to_noise_map_line.png" in plot_patch.paths


def test__subplot_ci_lines__is_output(ci_imaging_7x7, plot_path, plot_patch):

    aplt.CIImaging.subplot_ci_lines(
        ci_imaging=ci_imaging_7x7,
        line_region="parallel_front_edge",
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_ci_lines.png" in plot_patch.paths


def test__subplot_ci_imaging__is_output(ci_imaging_7x7, plot_path, plot_patch):

    aplt.CIImaging.subplot_ci_imaging(
        ci_imaging=ci_imaging_7x7,
        sub_plotter=aplt.SubPlotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/subplot_ci_imaging.png" in plot_patch.paths


def test__ci_individuals__output_dependent_on_inputs(
    ci_imaging_7x7, plot_path, plot_patch
):
    aplt.CIImaging.individual(
        ci_imaging=ci_imaging_7x7,
        plot_image=True,
        plot_ci_pre_cti=True,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/image.png" in plot_patch.paths

    assert f"{plot_path}/noise_map.png" not in plot_patch.paths

    assert f"{plot_path}/ci_pre_cti.png" in plot_patch.paths

    assert f"{plot_path}/signal_to_noise_map.png" not in plot_patch.paths


def test__ci_line_individuals__output_dependent_on_inputs(
    ci_imaging_7x7, plot_path, plot_patch
):

    aplt.CIImaging.individual_ci_lines(
        ci_imaging=ci_imaging_7x7,
        line_region="parallel_front_edge",
        plot_image=True,
        plot_ci_pre_cti=True,
        plotter=aplt.Plotter(output=aplt.Output(plot_path, format="png")),
    )

    assert f"{plot_path}/image_line.png" in plot_patch.paths

    assert f"{plot_path}/noise_map_line.png" not in plot_patch.paths

    assert f"{plot_path}/ci_pre_cti_line.png" in plot_patch.paths

    assert f"{plot_path}/signal_to_noise_map_line.png" not in plot_patch.paths



from autoarray.plot import plotters
from autocti.plot import ci_line_plots


@plotters.set_include_and_sub_plotter
@plotters.set_subplot_filename
def subplot_ci_imaging(ci_imaging, include=None, sub_plotter=None):
    """Plot the imaging data_type as a sub-plotters of all its quantites (e.g. the dataset, noise_map, PSF, Signal-to_noise-map, \
     etc).

    Set *autocti.data_type.array.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ci_imaging : data_type.ImagingData
        The imaging data_type, which includes the observed data_type, noise_map, PSF, signal-to-noise_map, etc.
    origin : True
        If true, the origin of the dataset's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : np.ndarray or data_type.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the dataset, this plots those pixels \
        over the immage.
    ignore_config : bool
        If `False`, the config file general.ini is used to determine whether the subpot is plotted. If `True`, the \
        config file is ignored.
    """

    number_subplots = 4

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    image(ci_imaging=ci_imaging, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    noise_map(ci_imaging=ci_imaging, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=3)

    ci_pre_cti(ci_imaging=ci_imaging, include=include, plotter=sub_plotter)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=4)

    signal_to_noise_map(ci_imaging=ci_imaging, include=include, plotter=sub_plotter)

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def individual(
    ci_imaging,
    plot_image=False,
    plot_noise_map=False,
    plot_signal_to_noise_map=False,
    plot_ci_pre_cti=False,
    plot_cosmic_ray_map=False,
    include=None,
    plotter=None,
):
    """Plot each attribute of the imaging data_type as individual figures one by one (e.g. the dataset, noise_map, PSF, \
     Signal-to_noise-map, etc).

    Set *autocti.data_type.array.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ci_imaging : data_type.ImagingData
        The imaging data_type, which includes the observed data_type, noise_map, PSF, signal-to-noise_map, etc.
    origin : True
        If true, the origin of the dataset's coordinate system is plotted as a 'x'.
    """

    if plot_image:

        image(ci_imaging=ci_imaging, include=include, plotter=plotter)

    if plot_noise_map:

        noise_map(ci_imaging=ci_imaging, include=include, plotter=plotter)

    if plot_signal_to_noise_map:

        signal_to_noise_map(ci_imaging=ci_imaging, include=include, plotter=plotter)

    if plot_ci_pre_cti:

        ci_pre_cti(ci_imaging=ci_imaging, include=include, plotter=plotter)

    if plot_cosmic_ray_map:

        cosmic_ray_map(ci_imaging=ci_imaging, include=include, plotter=plotter)


@plotters.set_include_and_plotter
@plotters.set_labels
def image(ci_imaging, include=None, plotter=None):
    """Plot the observed data_type of the imaging data_type.

    Set *autocti.data_type.array.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : data_type.ImagingData
        The imaging data_type, which includes the observed data_type, noise_map, PSF, signal-to-noise_map, etc.
    include_origin : True
        If true, the include_origin of the dataset's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : np.ndarray or data_type.array.grid_stacks.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the dataset, this plots those pixels \
        over the immage.
    """

    plotter.plot_frame(
        frame=ci_imaging.image,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def noise_map(ci_imaging, include=None, plotter=None):
    """Plot the noise_map of the imaging data_type.

    Set *autocti.data_type.array.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : data_type.ImagingData
        The imaging data_type, which includes the observed data_type, noise_map, PSF, signal-to-noise_map, etc.
    include_origin : True
        If true, the include_origin of the dataset's coordinate system is plotted as a 'x'.
    """
    plotter.plot_frame(
        frame=ci_imaging.noise_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def signal_to_noise_map(ci_imaging, include=None, plotter=None):
    """Plot the signal-to-noise_map of the imaging data_type.

    Set *autocti.data_type.array.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    image : data_type.ImagingData
        The imaging data_type, which includes the observed image, noise_map, PSF, signal-to-noise_map, etc.
    include_origin : True
        If true, the include_origin of the dataset's coordinate system is plotted as a 'x'.
    """
    plotter.plot_frame(
        frame=ci_imaging.signal_to_noise_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def ci_pre_cti(ci_imaging, include=None, plotter=None):
    """Plot the observed ci_pre_cti of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_pre_cti : CIFrame
        The ci_pre_cti of the dataset.
    """
    plotter.plot_frame(
        frame=ci_imaging.ci_pre_cti,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def cosmic_ray_map(ci_imaging, include=None, plotter=None):
    """Plot the observed ci_pre_cti of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_pre_cti : CIFrame
        The ci_pre_cti of the dataset.
    """
    plotter.plot_frame(
        frame=ci_imaging.cosmic_ray_map,
        include_origin=include.origin,
        include_parallel_overscan=include.parallel_overscan,
        include_serial_prescan=include.serial_prescan,
        include_serial_overscan=include.serial_overscan,
    )


@plotters.set_include_and_sub_plotter
@plotters.set_subplot_filename
def subplot_ci_lines(ci_imaging, line_region, include=None, sub_plotter=None):
    """Plot the ci simulator as a sub-plotters of all its quantites (e.g. the dataset, noise_map, PSF, Signal-to_noise-map, \
     etc).

    Set *autocti.simulator.arrays.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ci_imaging : simulator.CCDData
        The ci simulator, which includes the observed dataset, noise_map, PSF, signal-to-noise_map, etc.
    origin : True
        If true, the origin of the dataset's coordinate system is plotted as a 'x'.
    image_plane_pix_grid : np.ndarray or simulator.arrays.grid_lines.PixGrid
        If an adaptive pixelization whose pixels are formed by tracing pixels from the dataset, this plots those pixels \
        over the immage.
    ignore_config : bool
        If `False`, the config file general.ini is used to determine whether the subpot is plotted. If `True`, the \
        config file is ignored.
    """

    number_subplots = 4

    sub_plotter.open_subplot_figure(number_subplots=number_subplots)

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=1)

    image_line(
        ci_imaging=ci_imaging,
        line_region=line_region,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=2)

    noise_map_line(
        ci_imaging=ci_imaging,
        line_region=line_region,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=3)

    ci_pre_cti_line(
        ci_imaging=ci_imaging,
        line_region=line_region,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.setup_subplot(number_subplots=number_subplots, subplot_index=4)

    signal_to_noise_map_line(
        ci_imaging=ci_imaging,
        line_region=line_region,
        include=include,
        plotter=sub_plotter,
    )

    sub_plotter.output.subplot_to_figure()

    sub_plotter.figure.close()


def individual_ci_lines(
    ci_imaging,
    line_region,
    plot_image=False,
    plot_noise_map=False,
    plot_ci_pre_cti=False,
    plot_signal_to_noise_map=False,
    include=None,
    plotter=None,
):
    """Plot each attribute of the ci simulator as individual figures one by one (e.g. the dataset, noise_map, PSF, \
     Signal-to_noise-map, etc).

    Set *autocti.simulator.arrays.plotters.plotters* for a description of all innput parameters not described below.

    Parameters
    -----------
    ci_imaging : simulator.CCDData
        The ci simulator, which includes the observed dataset, noise_map, PSF, signal-to-noise_map, etc.
    origin : True
        If true, the origin of the dataset's coordinate system is plotted as a 'x'.
    """

    if plot_image:
        image_line(
            ci_imaging=ci_imaging,
            line_region=line_region,
            include=include,
            plotter=plotter,
        )

    if plot_noise_map:
        noise_map_line(
            ci_imaging=ci_imaging,
            line_region=line_region,
            include=include,
            plotter=plotter,
        )

    if plot_ci_pre_cti:
        ci_pre_cti_line(
            ci_imaging=ci_imaging,
            line_region=line_region,
            include=include,
            plotter=plotter,
        )

    if plot_signal_to_noise_map:
        signal_to_noise_map_line(
            ci_imaging=ci_imaging,
            line_region=line_region,
            include=include,
            plotter=plotter,
        )


@plotters.set_include_and_plotter
@plotters.set_labels
def image_line(ci_imaging, line_region, include=None, plotter=None):
    """Plot the observed image of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_imaging : CIFrame
        The image of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=ci_imaging.image,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def noise_map_line(ci_imaging, line_region, include=None, plotter=None):
    """Plot the observed noise_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_imaging : CIFrame
        The noise_map of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=ci_imaging.noise_map,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def ci_pre_cti_line(ci_imaging, line_region, include=None, plotter=None):
    """Plot the observed ci_pre_cti of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_imaging : CIFrame
        The ci_pre_cti of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=ci_imaging.ci_pre_cti,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )


@plotters.set_include_and_plotter
@plotters.set_labels
def signal_to_noise_map_line(ci_imaging, line_region, include=None, plotter=None):
    """Plot the observed signal_to_noise_map of the ccd simulator.

    Set *autocti.simulator.plotters.plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    ci_imaging : CIFrame
        The signal_to_noise_map of the dataset.
    """
    ci_line_plots.plot_line_from_ci_frame(
        ci_frame=ci_imaging.signal_to_noise_map,
        line_region=line_region,
        include=include,
        plotter=plotter,
    )



import os
import shutil

import numpy as np
import pytest
import autocti as ac
from autocti import exc


test_data_path = "{}/files/array/".format(os.path.dirname(os.path.realpath(__file__)))


class TestAPI:
    def test__manual__makes_array_with_pixel_scale(self):

        arr = ac.Array.manual_2d(array=[[1.0, 2.0], [3.0, 4.0]], pixel_scales=1.0)

        assert isinstance(arr, ac.Array)
        assert (arr == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)

        arr = ac.Array.manual_2d(
            array=[[1.0, 2.0], [3.0, 4.0]], pixel_scales=1.0, origin=(0.0, 1.0)
        )

        assert isinstance(arr, ac.Array)
        assert (arr == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

        arr = ac.Array.manual_2d(
            array=[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], pixel_scales=(2.0, 3.0)
        )

        assert isinstance(arr, ac.Array)
        assert (arr == np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]])).all()
        assert arr.pixel_scales == (2.0, 3.0)

    def test__manual_mask__makes_array_with_pixel_scale(self):
        mask = ac.Mask2D.unmasked(shape_2d=(2, 2), pixel_scales=1.0)
        arr = ac.Array.manual_mask(array=[[1.0, 2.0], [3.0, 4.0]], mask=mask)

        assert (arr == np.array([[1.0, 2.0], [3.0, 4.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)

        mask = ac.Mask2D.manual(
            mask=[[False, False], [True, False]], pixel_scales=1.0, origin=(0.0, 1.0)
        )
        arr = ac.Array.manual_mask(array=[[1.0, 2.0], [3.0, 4.0]], mask=mask)

        assert (arr == np.array([[1.0, 2.0], [0.0, 4.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

    def test__manual_2d__exception_raised_if_input_array_is_2d_and_not_shape_of_mask(
        self
    ):
        with pytest.raises(exc.ArrayException):
            mask = ac.Mask2D.unmasked(shape_2d=(2, 2), pixel_scales=1.0)
            ac.Array.manual_mask(array=[[1.0], [3.0]], mask=mask)

        with pytest.raises(exc.ArrayException):
            mask = ac.Mask2D.unmasked(shape_2d=(2, 2), pixel_scales=1.0)
            ac.Array.manual_mask(array=[[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]], mask=mask)

    def test__full__makes_scaled_array_with_pixel_scale(self):

        arr = ac.Array.ones(shape_2d=(2, 2), pixel_scales=1.0)

        assert isinstance(arr, ac.Array)
        assert (arr == np.array([[1.0, 1.0], [1.0, 1.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)

        arr = ac.Array.full(
            fill_value=2.0, shape_2d=(2, 2), pixel_scales=1.0, origin=(0.0, 1.0)
        )

        assert isinstance(arr, ac.Array)
        assert (arr == np.array([[2.0, 2.0], [2.0, 2.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

    def test__ones_zeros__makes_scaled_array_with_pixel_scale(self):

        arr = ac.Array.ones(shape_2d=(2, 2), pixel_scales=1.0)

        assert isinstance(arr, ac.Array)
        assert (arr == np.array([[1.0, 1.0], [1.0, 1.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)

        arr = ac.Array.zeros(shape_2d=(2, 2), pixel_scales=1.0, origin=(0.0, 1.0))

        assert isinstance(arr, ac.Array)
        assert (arr == np.array([[0.0, 0.0], [0.0, 0.0]])).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

    def test__from_fits__makes_array_without_other_inputs(self):

        arr = ac.Array.from_fits(
            file_path=test_data_path + "3x3_ones.fits", hdu=0, pixel_scales=1.0
        )

        assert isinstance(arr, ac.Array)
        assert (arr == np.ones((3, 3))).all()

        arr = ac.Array.from_fits(
            file_path=test_data_path + "4x3_ones.fits", hdu=0, pixel_scales=1.0
        )

        assert isinstance(arr, ac.Array)
        assert (arr == np.ones((4, 3))).all()

    def test__from_fits__makes_scaled_array_with_pixel_scale(self):

        arr = ac.Array.from_fits(
            file_path=test_data_path + "3x3_ones.fits", hdu=0, pixel_scales=1.0
        )

        assert isinstance(arr, ac.Array)
        assert (arr == np.ones((3, 3))).all()
        assert arr.pixel_scales == (1.0, 1.0)

        arr = ac.Array.from_fits(
            file_path=test_data_path + "4x3_ones.fits",
            hdu=0,
            pixel_scales=1.0,
            origin=(0.0, 1.0),
        )

        assert isinstance(arr, ac.Array)
        assert (arr == np.ones((4, 3))).all()
        assert arr.pixel_scales == (1.0, 1.0)
        assert arr.origin == (0.0, 1.0)

    def test__output_to_fits(self):

        arr = ac.Array.from_fits(
            file_path=test_data_path + "3x3_ones.fits", hdu=0, pixel_scales=1.0
        )

        output_data_dir = "{}/files/array/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        arr.output_to_fits(file_path=output_data_dir + "array.fits")

        array_from_out = ac.Array.from_fits(
            file_path=output_data_dir + "array.fits", hdu=0, pixel_scales=1.0
        )

        assert (array_from_out == np.ones((3, 3))).all()

    def test__output_to_fits__shapes_of_arrays_are_2d(self):

        arr = ac.Array.from_fits(
            file_path=test_data_path + "3x3_ones.fits", hdu=0, pixel_scales=1.0
        )

        output_data_dir = "{}/files/array/output_test/".format(
            os.path.dirname(os.path.realpath(__file__))
        )
        if os.path.exists(output_data_dir):
            shutil.rmtree(output_data_dir)

        os.makedirs(output_data_dir)

        arr.output_to_fits(file_path=output_data_dir + "array.fits")

        array_from_out = ac.util.array.numpy_array_2d_from_fits(
            file_path=output_data_dir + "array.fits", hdu=0
        )

        assert (array_from_out == np.ones((3, 3))).all()



from autoarray.fit.fit import FitImaging
from autoarray.structures.region import Region
from autoarray.structures.arrays.abstract_array import ExposureInfo
from autoarray.structures.frames.abstract_frame import Scans
from autoarray.structures.lines.lines import Line
from autoarray.structures.lines.lines import LineCollection
from autoarray.instruments import euclid
from autoarray.instruments import acs
from autoarray.dataset import preprocess

from .structures.arrays import Array
from .structures.frames import Frame
from .mask.mask import Mask2D
from .mask.mask import SettingsMask
from .dataset.imaging import Imaging
from .dataset.imaging import MaskedImaging
from .dataset.warm_pixels import find_warm_pixels
from . import charge_injection as ci
from .pipeline.phase.settings import SettingsCTI
from .pipeline.phase.settings import SettingsPhaseCIImaging
from .pipeline.phase.extensions import CombinedHyperPhase
from .pipeline.phase.ci_imaging.phase import PhaseCIImaging
from .pipeline.pipeline import Pipeline
from .util.clocker import Clocker
from .util.ccd import CCD
from .util.traps import TrapInstantCapture
from . import util
from . import plot

__version__ = '0.12.2'



import logging

from autofit.tools import pipeline as pl

logger = logging.getLogger(__name__)


class Pipeline(pl.Pipeline):
    def run(self, datasets, clocker, info=None, pool=None):
        def runner(phase, results):
            return phase.run(
                datasets=datasets,
                results=results,
                clocker=clocker,
                info=info,
                pool=pool,
            )

        return self.run_function(runner)



from autocti import exc


def plot_line_from_ci_frame(ci_frame, line_region, include=None, plotter=None):

    if line_region is "parallel_front_edge":
        line = ci_frame.parallel_front_edge_line_binned_over_columns()
    elif line_region is "parallel_trails":
        line = ci_frame.parallel_trails_line_binned_over_columns()
    elif line_region is "serial_front_edge":
        line = ci_frame.serial_front_edge_line_binned_over_rows()
    elif line_region is "serial_trails":
        line = ci_frame.serial_trails_line_binned_over_rows()
    else:
        raise exc.PlottingException(
            "The line region specified for the plotting of a line was invalid"
        )

    plotter.plot_line(y=line, x=range(len(line)))



import autocti as ac
import autofit as af
from test_autocti.integration.tests import runner

test_type = "parallel_and_serial"
test_name = "x1_species__x2_image__cosmic_rays"
ci_data_type = "ci_uniform_cosmic_rays"
ci_data_model = "parallel_x1__serial_x1"
resolution = "patch"
ci_normalizations = [10000.0, 84700.0]

clocker = ac.Clocker(parallel_express=2, serial_express=2)


def make_pipeline(name, folders, search=af.DynestyStatic()):

    parallel_ccd = af.PriorModel(ac.CCD)

    parallel_ccd.full_well_depth = 8.47e4
    parallel_ccd.well_notch_depth = 1e-7

    serial_ccd = af.PriorModel(ac.CCD)

    serial_ccd.full_well_depth = 8.47e4
    serial_ccd.well_notch_depth = 1e-7

    phase1 = ac.PhaseCIImaging(
        name="phase_1",
        folders=folders,
        search=search,
        parallel_traps=[af.PriorModel(ac.TrapInstantCapture)],
        parallel_ccd=parallel_ccd,
        serial_traps=[af.PriorModel(ac.TrapInstantCapture)],
        serial_ccd=serial_ccd,
    )

    return ac.Pipeline(name, phase1)


if __name__ == "__main__":

    import sys

    runner.run(sys.modules[__name__], clocker=clocker, load_cosmic_ray_map=True)



from copy import deepcopy

import numpy as np
from autoarray.structures import abstract_structure
from autoarray.structures.arrays import abstract_array
from autocti.charge_injection import ci_mask
from autocti.mask.mask import Mask2D
from autoarray.structures.frames import abstract_frame
from autoarray.instruments import euclid
from autoarray.util import array_util, frame_util


class AbstractCIFrame(abstract_frame.AbstractFrame):
    def __new__(
        cls,
        array,
        mask,
        ci_pattern,
        original_roe_corner=(1, 0),
        exposure_info=None,
        scans=None,
    ):
        """
        Class which represents the CCD quadrant of a charge injection image (e.g. the location of the parallel and
        serial front edge, trails).

        frame_geometry : CIFrame.CIQuadGeometry
            The quadrant geometry of the image, defining where the parallel / serial overscans are and
            therefore the direction of clocking and rotations before input into the cti algorithm.
        ci_pattern : CIPattern.CIPattern
            The charge injection ci_pattern (scans, normalization, etc.) of the charge injection image.
        """

        if type(array) is list:
            array = np.asarray(array)

        array[mask == True] = 0.0

        obj = array.view(cls)
        obj.mask = mask
        obj.store_in_1d = False
        obj.zoom_for_plot = False
        obj.original_roe_corner = original_roe_corner
        obj.exposure_info = exposure_info
        obj.scans = scans or abstract_frame.Scans()
        obj.ci_pattern = ci_pattern

        return obj

    def __array_finalize__(self, obj):

        super(AbstractCIFrame, self).__array_finalize__(obj)

        if isinstance(obj, AbstractCIFrame):
            if hasattr(obj, "mask"):
                self.mask = obj.mask
            if hasattr(obj, "original_roe_corner"):
                self.original_roe_corner = obj.original_roe_corner
            if hasattr(obj, "ci_pattern"):
                self.ci_pattern = obj.ci_pattern
            if hasattr(obj, "scans"):
                self.scans = obj.scans

    def _new_structure(self, array, mask, store_in_1d):
        return self.__class__(
            array=array,
            mask=mask,
            ci_pattern=self.ci_pattern,
            original_roe_corner=self.original_roe_corner,
            scans=self.scans,
            exposure_info=self.exposure_info,
        )

    @property
    def ci_regions_frame(self):
        """Extract an arrays of all of the charge-injection scans from a charge injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
          [...][ccccccccccccccccccccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the charge injection region and replaces all other values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][000]
          [000][ccccccccccccccccccccc][000]
        | [000][ccccccccccccccccccccc][000]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][000000000000000000000][000]    | of
        | [000][ccccccccccccccccccccc][000]    | clocking
          [000][ccccccccccccccccccccc][000]    |

        []     [=====================]
               <---------S----------



        """

        new_array = self.copy() * 0.0

        for region in self.ci_pattern.regions:
            new_array[region.slice] += self[region.slice]

        return new_array

    @property
    def non_ci_regions_frame(self):
        """Extract an arrays of all of the parallel trails following the charge-injection scans from a charge
        injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [tptpptptptpptpptpptpt]
               [tptptptpptpttptptptpt]
          [...][ttttttttttttttttttttt][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
          [...][ccccccccccccccccccccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [tptpptptptpptpptpptpt]
               [tptptptpptpttptptptpt]
          [000][ttttttttttttttttttttt][000]
          [000][000000000000000000000][000]
        | [000][000000000000000000000][000]    |
        | [000][ttttttttttttttttttttt][000]    | Direction
        P [000][ttttttttttttttttttttt][000]    | of
        | [000][000000000000000000000][000]    | clocking
          [000][000000000000000000000][000]    |

        []     [=====================]
               <---------S----------
        """

        non_ci_regions_array = self.copy()

        for region in self.ci_pattern.regions:
            non_ci_regions_array[region.slice] = 0.0

        return non_ci_regions_array

    @property
    def parallel_non_ci_regions_frame(self):
        """Extract an arrays of all of the parallel trails following the charge-injection scans from a charge
        injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [tptpptptptpptpptpptpt]
               [tptptptpptpttptptptpt]
          [...][ttttttttttttttttttttt][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
          [...][ccccccccccccccccccccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [tptpptptptpptpptpptpt]
               [tptptptpptpttptptptpt]
          [000][ttttttttttttttttttttt][000]
          [000][000000000000000000000][000]
        | [000][000000000000000000000][000]    |
        | [000][ttttttttttttttttttttt][000]    | Direction
        P [000][ttttttttttttttttttttt][000]    | of
        | [000][000000000000000000000][000]    | clocking
          [000][000000000000000000000][000]    |

        []     [=====================]
               <---------S----------
        """

        parallel_frame = self.non_ci_regions_frame

        parallel_frame[self.scans.serial_prescan.slice] = 0.0
        parallel_frame[self.scans.serial_overscan.slice] = 0.0

        return parallel_frame

    def parallel_edges_and_trails_frame(self, front_edge_rows=None, trails_rows=None):
        """Extract an arrays of all of the parallel front edges and trails of each the charge-injection scans from
        a charge injection ci_frame.

        One can specify the range of rows that are extracted, for example:

        front_edge_rows = (0, 1) will extract just the leading front edge row.
        front_edge_rows = (0, 2) will extract the leading two front edge rows.
        trails_rows = (0, 1) will extract the first row of trails closest to the charge injection region.

        The diagram below illustrates the arrays that is extracted from a ci_frame for front_edge_rows=(0,1) and
        trails_rows=(0,1):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [tptpptptptpptpptpptpt]
               [tptptptpptpttptptptpt]
          [...][ttttttttttttttttttttt][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
          [...][ccccccccccccccccccccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the leading edges and trails following all charge injection scans and
        replaces all other values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][ttttttttttttttttttttt][000]
          [000][000000000000000000000][000]
        | [000][ccccccccccccccccccccc][000]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][ttttttttttttttttttttt][000]    | of
        | [000][000000000000000000000][000]    | clocking
          [000][ccccccccccccccccccccc][000]    |

        []     [=====================]
               <---------S----------

        Parameters
        ------------
        array
        front_edge_rows : (int, int)
            The row indexes to extract the front edge between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows).
        trails_rows : (int, int)
            The row indexes to extract the trails between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        new_frame = self.copy() * 0.0

        if front_edge_rows is not None:

            front_regions = list(
                map(
                    lambda ci_region: self.parallel_front_edge_of_region(
                        ci_region, front_edge_rows
                    ),
                    self.ci_pattern.regions,
                )
            )

            front_edges = self.parallel_front_edge_arrays(rows=front_edge_rows)

            for i, region in enumerate(front_regions):
                new_frame[region.y0 : region.y1, region.x0 : region.x1] += front_edges[
                    i
                ]

        if trails_rows is not None:

            trails_regions = list(
                map(
                    lambda ci_region: self.parallel_trails_of_region(
                        ci_region, trails_rows
                    ),
                    self.ci_pattern.regions,
                )
            )

            trails = self.parallel_trails_arrays(rows=trails_rows)

            for i, region in enumerate(trails_regions):
                new_frame[region.y0 : region.y1, region.x0 : region.x1] += trails[i]

        return new_frame

    def parallel_calibration_frame_from_columns(self, columns):
        """Extract an parallel calibration array from a charge injection ci_frame, where this arrays is a sub-set of
        the ci_frame which be used for just parallel calibration. Specifically, this ci_frame is a specified number
        of columns closest to the read-out electronics.

        The diagram below illustrates the arrays that is extracted from a ci_frame with columns=(0, 3):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / parallel charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ptptptptptptptptptptp]
               [tptptptptptptptptptpt]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][sss]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][sss]    | clocking
          [...][ccccccccccccccccccccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [ptp]
               [tpt]
               [xxx]
               [ccc]
        |      [ccc]                           |
        |      [xxx]                           | Direction
        P      [xxx]                           | of
        |      [ccc]                           | clocking
               [ccc]                           |

        []     [=====================]
               <---------S----------
        """
        extraction_region = self.parallel_side_nearest_read_out_region(
            region=self.ci_pattern.regions[0], columns=columns
        )
        return CIFrame.extracted_ci_frame_from_ci_frame_and_extraction_region(
            ci_frame=self, extraction_region=extraction_region
        )

    def parallel_calibration_mask_from_mask_and_columns(self, mask, columns):
        extraction_region = self.parallel_side_nearest_read_out_region(
            region=self.ci_pattern.regions[0], columns=columns
        )
        return ci_mask.CIMask(
            mask=mask[extraction_region.slice], pixel_scales=mask.pixel_scales
        )

    @property
    def serial_trails_frame(self):
        """Extract an arrays of all of the serial trails in the serial overscan region, that are to the side of a
        charge-injection scans from a charge injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][000]
          [000][000000000000000000000][tst]
        | [000][000000000000000000000][sts]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][000000000000000000000][000]    | of
        | [000][000000000000000000000][tst]    | clocking
          [000][000000000000000000000][sts]    |

        []     [=====================]
               <---------S----------
        """
        array = self.serial_edges_and_trails_frame(
            trails_columns=(0, self.scans.serial_overscan.total_columns)
        )
        return array

    @property
    def serial_overscan_no_trails_frame(self):
        """Extract an arrays of all of the scans of the serial overscan that don't contain trails from a
        charge injection region (i.e. are not to the side of one).

        The diagram below illustrates the arrays that is extracted from a ci_frame:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][sss]
          [000][000000000000000000000][000]
        | [000][000000000000000000000][000]    |
        | [000][000000000000000000000][sss]    | Direction
        P [000][000000000000000000000][sss]    | of
        | [000][000000000000000000000][000]    | clocking
          [000][000000000000000000000][000]    |

        []     [=====================]
               <---------S----------
        """
        new_array = self.copy() * 0.0
        overscan_slice = self.scans.serial_overscan.slice

        new_array[overscan_slice] = self[overscan_slice]

        trails_regions = list(
            map(
                lambda ci_region: self.serial_trails_of_region(
                    ci_region, (0, self.scans.serial_overscan.total_columns)
                ),
                self.ci_pattern.regions,
            )
        )

        for region in trails_regions:
            new_array[region.slice] = 0

        return new_array

    def serial_edges_and_trails_frame(
        self, front_edge_columns=None, trails_columns=None
    ):
        """Extract an arrays of all of the serial front edges and trails of each the charge-injection scans from
        a charge injection ci_frame.

        One can specify the range of columns that are extracted, for example:

        front_edge_columns = (0, 1) will extract just the leading front edge column.
        front_edge_columns = (0, 2) will extract the leading two front edge columns.
        trails_columns = (0, 1) will extract the first column of trails closest to the charge injection region.

        The diagram below illustrates the arrays that is extracted from a ci_frame for front_edge_columns=(0,2) and
        trails_columns=(0,2):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][sts]
        | [...][ccccccccccccccccccccc][tst]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][sts]    | clocking
          [...][ccccccccccccccccccccc][tst]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the leading edge and trails following all charge injection scans and
        replaces all other values with 0s:

               [000000000000000000000]
               [000000000000000000000]
          [000][000000000000000000000][000]
          [000][cc0000000000000000000][st0]
        | [000][cc0000000000000000000][ts0]    |
        | [000][000000000000000000000][000]    | Direction
        P [000][000000000000000000000][000]    | of
        | [000][cc0000000000000000000][st0]    | clocking
          [000][cc0000000000000000000][st0]    |

        []     [=====================]
               <---------S----------

        Parameters
        ------------
        array
        front_edge_columns : (int, int)
            The column indexes to extract the front edge between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd rows)
        trails_columns : (int, int)
            The column indexes to extract the trails between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        new_array = self.copy() * 0.0

        if front_edge_columns is not None:

            front_regions = list(
                map(
                    lambda ci_region: self.serial_front_edge_of_region(
                        ci_region, front_edge_columns
                    ),
                    self.ci_pattern.regions,
                )
            )

            front_edges = self.serial_front_edge_arrays(columns=front_edge_columns)

            for i, region in enumerate(front_regions):
                new_array[region.y0 : region.y1, region.x0 : region.x1] += front_edges[
                    i
                ]

        if trails_columns is not None:

            trails_regions = list(
                map(
                    lambda ci_region: self.serial_trails_of_region(
                        ci_region, trails_columns
                    ),
                    self.ci_pattern.regions,
                )
            )

            trails = self.serial_trails_arrays(columns=trails_columns)

            for i, region in enumerate(trails_regions):
                new_array[region.y0 : region.y1, region.x0 : region.x1] += trails[i]

        return new_array

    def serial_calibration_frame_from_rows(self, rows):
        """Extract a serial calibration array from a charge injection ci_frame, where this arrays is a sub-set of the
        ci_frame which can be used for serial-only calibration. Specifically, this ci_frame is all charge injection
        scans and their serial over-scan trails.

        The diagram below illustrates the arrays that is extracted from a ci_frame with column=5:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [pppppppppppppppppppp ]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

        |                                      |
        |      [cccccccccccccccc][tst]         | Direction
        P      [cccccccccccccccc][tst]         | of
        |      [cccccccccccccccc][tst]         | clocking
               [cccccccccccccccc][tst]         |

        []     [=====================]
               <---------S----------
        """
        calibration_images = self.serial_calibration_sub_arrays
        calibration_images = list(
            map(lambda image: image[rows[0] : rows[1], :], calibration_images)
        )

        array = np.concatenate(calibration_images, axis=0)

        # TODO : can we generalize this method for multiple extracts? Feels too complicated so just doing it for this
        # TODO : specific case for now.

        serial_prescan = (
            (
                0,
                array.shape[0],
                self.scans.serial_prescan[2],
                self.scans.serial_prescan[3],
            )
            if self.scans.serial_prescan is not None
            else None
        )
        serial_overscan = (
            (
                0,
                array.shape[0],
                self.scans.serial_overscan[2],
                self.scans.serial_overscan[3],
            )
            if self.scans.serial_overscan is not None
            else None
        )

        x0 = self.ci_pattern.regions[0][2]
        x1 = self.ci_pattern.regions[0][3]
        offset = 0
        new_ci_pattern_regions = []
        for region in self.ci_pattern.regions:
            ysize = rows[1] - rows[0]
            new_ci_pattern_regions.append((offset, offset + ysize, x0, x1))
            offset += ysize

        new_ci_pattern = deepcopy(self.ci_pattern)
        new_ci_pattern.regions = new_ci_pattern_regions

        return CIFrame.manual(
            array=array,
            ci_pattern=new_ci_pattern,
            roe_corner=self.original_roe_corner,
            scans=abstract_frame.Scans(
                parallel_overscan=None,
                serial_prescan=serial_prescan,
                serial_overscan=serial_overscan,
            ),
            pixel_scales=self.pixel_scales,
        )

    def serial_calibration_mask_from_mask_and_rows(self, mask, rows):
        """Extract a serial calibration array from a charge injection ci_frame, where this arrays is a sub-set of the
        ci_frame which can be used for serial-only calibration. Specifically, this ci_frame is all charge injection
        scans and their serial over-scan trails.

        The diagram below illustrates the arrays that is extracted from a ci_frame with column=5:

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan    [cccccccccc] = charge injection region
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [pppppppppppppppppppp ]
          [...][xxxxxxxxxxxxxxxxxxxxx][sss]
          [...][ccccccccccccccccccccc][tst]
        | [...][ccccccccccccccccccccc][sts]    |
        | [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | Direction
        P [...][xxxxxxxxxxxxxxxxxxxxx][sss]    | of
        | [...][ccccccccccccccccccccc][tst]    | clocking
          [...][ccccccccccccccccccccc][sts]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans and replaces all other
        values with 0s:

        |                                      |
        |      [cccccccccccccccc][tst]         | Direction
        P      [cccccccccccccccc][tst]         | of
        |      [cccccccccccccccc][tst]         | clocking
               [cccccccccccccccc][tst]         |

        []     [=====================]
               <---------S----------
        """

        calibration_regions = list(
            map(
                lambda ci_region: self.serial_entire_rows_of_region(region=ci_region),
                self.ci_pattern.regions,
            )
        )
        calibration_masks = list(
            map(lambda region: mask[region.slice], calibration_regions)
        )

        calibration_masks = list(
            map(lambda mask: mask[rows[0] : rows[1], :], calibration_masks)
        )
        return ci_mask.CIMask(
            mask=np.concatenate(calibration_masks, axis=0),
            pixel_scales=mask.pixel_scales,
        )

    @property
    def serial_calibration_sub_arrays(self):
        """Extract each charge injection region image for the serial calibration arrays above."""

        calibration_regions = list(
            map(
                lambda ci_region: self.serial_entire_rows_of_region(region=ci_region),
                self.ci_pattern.regions,
            )
        )
        return list(map(lambda region: self[region.slice], calibration_regions))

    def parallel_front_edge_line_binned_over_columns(self, rows=None):
        front_stacked_array = self.parallel_front_edge_stacked_array(rows=rows)
        return np.ma.mean(np.ma.asarray(front_stacked_array), axis=1)

    def parallel_front_edge_stacked_array(self, rows=None):
        front_arrays = self.parallel_front_edge_arrays(rows=rows)
        return np.ma.mean(np.ma.asarray(front_arrays), axis=0)

    def parallel_front_edge_arrays(self, rows=None):
        """Extract a list of structures of the parallel front edge scans of a charge injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame for rows=(0, 1):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci_region index)
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][ttttttttttttttttttttt][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        | [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the front edges of all charge injection scans.

        list index 0:

        [c0c0c0cc0c0c0c0c0c0c0]

        list index 1:

        [1c1c1c1c1c1c1c1c1c1c1]

        Parameters
        ------------
        array
        rows : (int, int)
            The row indexes to extract the front edge between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        front_regions = self.parallel_front_edge_regions(rows=rows)
        front_arrays = list(map(lambda region: self[region.slice], front_regions))
        front_masks = list(map(lambda region: self.mask[region.slice], front_regions))
        front_arrays = list(
            map(
                lambda front_array, front_mask: np.ma.array(
                    front_array, mask=front_mask
                ),
                front_arrays,
                front_masks,
            )
        )
        return front_arrays

    def parallel_front_edge_regions(self, rows=None):
        """Calculate a list of the parallel front edge scans of a charge injection ci_frame.

        The diagram below illustrates the region that calculaed from a ci_frame for rows=(0, 1):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci_region index)
        [xxxxxxxxxx]                [tttttttttt] = parallel / serial charge injection region trail

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][ttttttttttttttttttttt][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        | [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the front edges of all charge injection scans.

        list index 0:

        [0, 1, 3, 21] (serial prescan is 3 pixels)

        list index 1:

        [3, 4, 3, 21] (serial prescan is 3 pixels)

        Parameters
        ------------
        arrays
        rows : (int, int)
            The row indexes to extract the front edge between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        if rows is None:
            rows = (0, self.ci_pattern.total_rows_min)
        return list(
            map(
                lambda ci_region: self.parallel_front_edge_of_region(
                    region=ci_region, rows=rows
                ),
                self.ci_pattern.regions,
            )
        )

    def parallel_trails_line_binned_over_columns(self, rows=None):
        trails_stacked_array = self.parallel_trails_stacked_array(rows=rows)
        return np.ma.mean(np.ma.asarray(trails_stacked_array), axis=1)

    def parallel_trails_stacked_array(self, rows=None):
        trails_arrays = self.parallel_trails_arrays(rows=rows)
        return np.ma.mean(np.ma.asarray(trails_arrays), axis=0)

    def parallel_trails_arrays(self, rows=None):
        """Extract the parallel trails of a charge injection ci_frame.


        The diagram below illustrates the arrays that is extracted from a ci_frame for rows=(0, 1):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci region index)
        [xxxxxxxxxx]
        [t#t#t#t#t#] = parallel / serial charge injection region trail (0 / 1 indicates ci region index)

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][t1t1t1t1t1t1t1t1t1t1t][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        | [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        | [...][t0t0t0t0t0t0t0t0t0t0t][sss]    | Direction
        P [...][0t0t0t0t0t0t0t0t0t0t0][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans:

        list index 0:

        [t0t0t0tt0t0t0t0t0t0t0]

        list index 1:

        [1t1t1t1t1t1t1t1t1t1t1]

        Parameters
        ------------
        array
        rows : (int, int)
            The row indexes to extract the trails between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows)
        """
        trails_regions = self.parallel_trails_regions(rows=rows)
        trails_arrays = list(map(lambda region: self[region.slice], trails_regions))
        trails_masks = list(map(lambda region: self.mask[region.slice], trails_regions))
        trails_arrays = list(
            map(
                lambda trails_array, front_mask: np.ma.array(
                    trails_array, mask=front_mask
                ),
                trails_arrays,
                trails_masks,
            )
        )
        return trails_arrays

    def parallel_trails_regions(self, rows=None):
        """
    Returns the parallel scans of a charge injection ci_frame.

        The diagram below illustrates the region that is calculated from a ci_frame for rows=(0, 1):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci region index)
        [xxxxxxxxxx]
        [t#t#t#t#t#] = parallel / serial charge injection region trail (0 / 1 indicates ci region index)

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][t1t1t1t1t1t1t1t1t1t1t][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        | [...][1c1c1cc1c1cc1ccc1cc1][sss]    |
        | [...][t0t0t0t0t0t0t0t0t0t0t][sss]    | Direction
        P [...][0t0t0t0t0t0t0t0t0t0t0][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the trails following all charge injection scans:

        list index 0:

        [2, 4, 3, 21] (serial prescan is 3 pixels)

        list index 1:

        [6, 7, 3, 21] (serial prescan is 3 pixels)

        Parameters
        ------------
        arrays
        rows : (int, int)
            The row indexes to extract the trails between (e.g. rows(0, 3) extracts the 1st, 2nd and 3rd rows)
        """

        if rows is None:
            rows = (0, self.smallest_parallel_trails_rows_to_frame_edge)

        return list(
            map(
                lambda ci_region: self.parallel_trails_of_region(ci_region, rows),
                self.ci_pattern.regions,
            )
        )

    def serial_front_edge_line_binned_over_rows(self, columns=None):
        front_stacked_array = self.serial_front_edge_stacked_array(columns=columns)
        return np.ma.mean(np.ma.asarray(front_stacked_array), axis=0)

    def serial_front_edge_stacked_array(self, columns=None):
        front_arrays = self.serial_front_edge_arrays(columns=columns)
        return np.ma.mean(np.ma.asarray(front_arrays), axis=0)

    def serial_front_edge_arrays(self, columns=None):
        """Extract a list of the serial front edge structures of a charge injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame for columnss=(0, 3):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci_region index)
        [xxxxxxxxxx]
        [tttttttttt] = parallel / serial charge injection region trail ((0 / 1 indicates ci_region index)

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][ttttttttttttttttttttt][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        | [...][1c1c1cc1c1cc1ccc1cc1c][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the serial front edges of all charge injection scans.

        list index 0:

        [c0c0]

        list index 1:

        [1c1c]

        Parameters
        ------------
        array
        columns : (int, int)
            The column indexes to extract the front edge between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd
            columns)
        """
        front_regions = self.serial_front_edge_regions(columns=columns)
        front_arrays = list(map(lambda region: self[region.slice], front_regions))
        front_masks = list(map(lambda region: self.mask[region.slice], front_regions))
        front_arrays = list(
            map(
                lambda front_array, front_mask: np.ma.array(
                    front_array, mask=front_mask
                ),
                front_arrays,
                front_masks,
            )
        )
        return front_arrays

    def serial_front_edge_regions(self, columns=None):
        """
    Returns a list of the serial front edges scans of a charge injection ci_frame.

        The diagram below illustrates the region that is calculated from a ci_frame for columns=(0, 4):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci_region index)
        [xxxxxxxxxx]
        [tttttttttt] = parallel / serial charge injection region trail ((0 / 1 indicates ci_region index)

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][ttttttttttttttttttttt][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][sss]
        | [...][1c1c1cc1c1cc1ccc1cc1c][sss]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][sss]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][sss]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the serial front edges of all charge injection scans.

        list index 0:

        [0, 2, 3, 21] (serial prescan is 3 pixels)

        list index 1:

        [4, 6, 3, 21] (serial prescan is 3 pixels)

        Parameters
        ------------
        arrays
        columns : (int, int)
            The column indexes to extract the front edge between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd
            columns)
        """
        if columns is None:
            columns = (0, self.ci_pattern.total_columns_min)
        return list(
            map(
                lambda ci_region: self.serial_front_edge_of_region(ci_region, columns),
                self.ci_pattern.regions,
            )
        )

    def serial_trails_line_binned_over_rows(self, columns=None):
        trails_stacked_array = self.serial_trails_stacked_array(columns=columns)
        return np.ma.mean(np.ma.asarray(trails_stacked_array), axis=0)

    def serial_trails_stacked_array(self, columns=None):
        front_arrays = self.serial_trails_arrays(columns=columns)
        return np.ma.mean(np.ma.asarray(front_arrays), axis=0)

    def serial_trails_arrays(self, columns=None):
        """Extract a list of the serial trails of a charge injection ci_frame.

        The diagram below illustrates the arrays that is extracted from a ci_frame for columnss=(0, 3):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci_region index)
        [xxxxxxxxxx]
        [tttttttttt] = parallel / serial charge injection region trail ((0 / 1 indicates ci_region index)

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][ttttttttttttttttttttt][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][st1]
        | [...][1c1c1cc1c1cc1ccc1cc1c][ts0]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][st1]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][ts0]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the serial front edges of all charge injection scans.

        list index 0:

        [st0]

        list index 1:

        [st1]

        Parameters
        ------------
        array
        columns : (int, int)
            The column indexes to extract the trails between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd columns)
        """
        trails_regions = self.serial_trails_regions(columns=columns)
        trails_arrays = list(map(lambda region: self[region.slice], trails_regions))
        trails_masks = list(map(lambda region: self.mask[region.slice], trails_regions))
        trails_arrays = list(
            map(
                lambda trails_array, front_mask: np.ma.array(
                    trails_array, mask=front_mask
                ),
                trails_arrays,
                trails_masks,
            )
        )
        return trails_arrays

    def serial_trails_regions(self, columns=None):
        """
    Returns a list of the serial trails scans of a charge injection ci_frame.

        The diagram below illustrates the region is calculated from a ci_frame for columnss=(0, 4):

        ---KEY---
        ---------

        [] = read-out electronics   [==========] = read-out register

        [xxxxxxxxxx]                [..........] = serial prescan       [ssssssssss] = serial overscan
        [xxxxxxxxxx] = CCD panel    [pppppppppp] = parallel overscan
        [c#cc#c#c#c] = charge injection region (0 / 1 indicates ci_region index)
        [xxxxxxxxxx]
        [tttttttttt] = parallel / serial charge injection region trail ((0 / 1 indicates ci_region index)

        P = Parallel Direction      S = Serial Direction

               [ppppppppppppppppppppp]
               [ppppppppppppppppppppp]
          [...][ttttttttttttttttttttt][sss]
          [...][c1c1cc1c1cc1cc1ccc1cc][st1]
        | [...][1c1c1cc1c1cc1ccc1cc1c][ts0]    |
        | [...][ttttttttttttttttttttt][sss]    | Direction
        P [...][ttttttttttttttttttttt][sss]    | of
        | [...][0ccc0cccc0cccc0cccc0c][st1]    | clocking
          [...][cc0ccc0cccc0cccc0cccc][ts0]    |

        []     [=====================]
               <---------S----------

        The extracted ci_frame keeps just the serial front edges of all charge injection scans.

        list index 0:

        [0, 2, 22, 225 (serial prescan is 3 pixels)

        list index 1:

        [4, 6, 22, 25] (serial prescan is 3 pixels)

        Parameters
        ------------
        arrays
        columns : (int, int)
            The column indexes to extract the trails between (e.g. columns(0, 3) extracts the 1st, 2nd and 3rd columns)
        """
        if columns is None:
            columns = (0, self.scans.serial_trails_columns)
        return list(
            map(
                lambda ci_region: self.serial_trails_of_region(ci_region, columns),
                self.ci_pattern.regions,
            )
        )

    @property
    def parallel_serial_calibration_frame(self):
        return self

    @property
    def smallest_parallel_trails_rows_to_frame_edge(self):

        rows_between_regions = self.ci_pattern.rows_between_regions
        rows_between_regions.append(self.parallel_trail_size_to_frame_edge)
        return np.min(rows_between_regions)

    @property
    def parallel_trail_size_to_frame_edge(self):

        return self.shape_2d[0] - np.max(
            [region.y1 for region in self.ci_pattern.regions]
        )


class CIFrame(AbstractCIFrame):
    @classmethod
    def manual(
        cls,
        array,
        pixel_scales,
        ci_pattern,
        roe_corner=(1, 0),
        exposure_info=None,
        scans=None,
    ):
        """Abstract class for the geometry of a CTI Image.

        A FrameArray is stored as a 2D ndarrays. When this immage is passed to arctic, clocking goes towards
        the 'top' of the ndarrays (e.g. towards row 0). Trails therefore appear towards the 'bottom' of the arrays
        (e.g. the final row).

        Arctic has no in-built functionality for changing the direction of clocking depending on the input
        configuration file. Therefore, image rotations are handled before arctic is called, using the functions
        defined in this class (and its children). These routines define how an image is rotated before parallel
        and serial clocking and how to reorient the image back to its original orientation after clocking is performed.

        Currently, only four geometries are available, which are specific to Euclid (and documented in the
        *QuadGeometryEuclid* class).

        Parameters
        -----------
        parallel_overscan : frame.Region
            The parallel overscan region of the ci_frame.
        serial_prescan : frame.Region
            The serial prescan region of the ci_frame.
        serial_overscan : frame.Region
            The serial overscan region of the ci_frame.
        """

        if array is None:
            return None

        array = abstract_array.convert_array(array=array)

        pixel_scales = abstract_structure.convert_pixel_scales(
            pixel_scales=pixel_scales
        )

        mask = Mask2D.unmasked(shape_2d=array.shape, pixel_scales=pixel_scales)

        scans = abstract_frame.Scans.rotated_from_roe_corner(
            roe_corner=roe_corner, shape_2d=array.shape, scans=scans
        )

        return CIFrame(
            array=frame_util.rotate_array_from_roe_corner(
                array=array, roe_corner=roe_corner
            ),
            mask=mask,
            ci_pattern=frame_util.rotate_ci_pattern_from_roe_corner(
                ci_pattern=ci_pattern, shape_2d=array.shape, roe_corner=roe_corner
            ),
            original_roe_corner=roe_corner,
            exposure_info=exposure_info,
            scans=scans,
        )

    @classmethod
    def manual_mask(
        cls, array, mask, ci_pattern, roe_corner=(1, 0), exposure_info=None, scans=None
    ):
        """Abstract class for the geometry of a CTI Image.

        A FrameArray is stored as a 2D ndarrays. When this immage is passed to arctic, clocking goes towards
        the 'top' of the ndarrays (e.g. towards row 0). Trails therefore appear towards the 'bottom' of the arrays
        (e.g. the final row).

        Arctic has no in-built functionality for changing the direction of clocking depending on the input
        configuration file. Therefore, image rotations are handled before arctic is called, using the functions
        defined in this class (and its children). These routines define how an image is rotated before parallel
        and serial clocking and how to reorient the image back to its original orientation after clocking is performed.

        Currently, only four geometries are available, which are specific to Euclid (and documented in the
        *QuadGeometryEuclid* class).

        Parameters
        -----------
        parallel_overscan : Region
            The parallel overscan region of the ci_frame.
        serial_prescan : Region
            The serial prescan region of the ci_frame.
        serial_overscan : Region
            The serial overscan region of the ci_frame.
        """

        array = abstract_array.convert_array(array=array)

        array = frame_util.rotate_array_from_roe_corner(
            array=array, roe_corner=roe_corner
        )
        mask = frame_util.rotate_array_from_roe_corner(
            array=mask, roe_corner=roe_corner
        )

        array[mask == True] = 0.0

        scans = abstract_frame.Scans.rotated_from_roe_corner(
            roe_corner=roe_corner, shape_2d=array.shape, scans=scans
        )

        return CIFrame(
            array=array,
            mask=mask,
            ci_pattern=frame_util.rotate_ci_pattern_from_roe_corner(
                ci_pattern=ci_pattern, shape_2d=array.shape, roe_corner=roe_corner
            ),
            original_roe_corner=roe_corner,
            exposure_info=exposure_info,
            scans=scans,
        )

    @classmethod
    def full(
        cls,
        fill_value,
        shape_2d,
        pixel_scales,
        ci_pattern,
        roe_corner=(1, 0),
        exposure_info=None,
        scans=None,
    ):

        return cls.manual(
            array=np.full(fill_value=fill_value, shape=shape_2d),
            pixel_scales=pixel_scales,
            ci_pattern=ci_pattern,
            roe_corner=roe_corner,
            exposure_info=exposure_info,
            scans=scans,
        )

    @classmethod
    def ones(
        cls,
        shape_2d,
        pixel_scales,
        ci_pattern,
        roe_corner=(1, 0),
        exposure_info=None,
        scans=None,
    ):
        return cls.full(
            fill_value=1.0,
            shape_2d=shape_2d,
            pixel_scales=pixel_scales,
            ci_pattern=ci_pattern,
            roe_corner=roe_corner,
            exposure_info=exposure_info,
            scans=scans,
        )

    @classmethod
    def zeros(
        cls,
        shape_2d,
        pixel_scales,
        ci_pattern,
        roe_corner=(1, 0),
        exposure_info=None,
        scans=None,
    ):
        return cls.full(
            fill_value=0.0,
            shape_2d=shape_2d,
            pixel_scales=pixel_scales,
            ci_pattern=ci_pattern,
            roe_corner=roe_corner,
            exposure_info=exposure_info,
            scans=scans,
        )

    @classmethod
    def extracted_ci_frame_from_ci_frame_and_extraction_region(
        cls, ci_frame, extraction_region
    ):

        scans = abstract_frame.Scans.after_extraction(
            frame=ci_frame, extraction_region=extraction_region
        )

        return cls.manual(
            array=ci_frame[extraction_region.slice],
            ci_pattern=ci_frame.ci_pattern.with_extracted_regions(
                extraction_region=extraction_region
            ),
            exposure_info=ci_frame.exposure_info,
            scans=scans,
            roe_corner=ci_frame.original_roe_corner,
            pixel_scales=ci_frame.pixel_scales,
        )

    @classmethod
    def from_fits(
        cls,
        file_path,
        hdu,
        pixel_scales,
        ci_pattern,
        roe_corner=(1, 0),
        exposure_info=None,
        scans=None,
    ):
        """Load the image ci_data from a fits file.

        Params
        ----------
        path : str
            The path to the ci_data
        filename : str
            The file name of the fits image ci_data.
        hdu : int
            The HDU number in the fits file containing the image ci_data.
        frame_geometry : FrameArray.FrameGeometry
            The geometry of the ci_frame, defining the direction of parallel and serial clocking and the \
            locations of different scans of the CCD (overscans, prescan, etc.)
        """

        pixel_scales = abstract_structure.convert_pixel_scales(
            pixel_scales=pixel_scales
        )

        array = array_util.numpy_array_2d_from_fits(file_path=file_path, hdu=hdu)

        return CIFrame.manual(
            array=array,
            ci_pattern=ci_pattern,
            roe_corner=roe_corner,
            exposure_info=exposure_info,
            scans=scans,
            pixel_scales=pixel_scales,
        )

    @classmethod
    def from_ci_frame(cls, ci_frame, mask):
        return CIFrame(
            array=ci_frame,
            mask=mask,
            ci_pattern=ci_frame.ci_pattern,
            original_roe_corner=ci_frame.original_roe_corner,
            exposure_info=ci_frame.exposure_info,
            scans=abstract_frame.Scans.from_frame(frame=ci_frame),
        )


class CIFrameEuclid(CIFrame):
    @classmethod
    def from_ccd_and_quadrant_id(
        cls,
        array,
        ccd_id,
        quadrant_id,
        ci_pattern,
        parallel_size=2086,
        serial_size=2128,
        serial_prescan_size=51,
        serial_overscan_size=29,
        parallel_overscan_size=20,
    ):
        """Before reading this docstring, read the docstring for the __init__function above.

        In the Euclid FPA, the quadrant id ('E', 'F', 'G', 'H') depends on whether the CCD is located
        on the left side (rows 1-3) or right side (rows 4-6) of the FPA:

        LEFT SIDE ROWS 1-2-3
        --------------------

         <--------S-----------   ---------S----------->
        [] [========= 2 =========] [========= 3 =========] []          |
        /    [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  /          |
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | Direction arctic
        P   [xxxxxxxxx H xxxxxxxxx] [xxxxxxxxx G xxxxxxxxx]  P         | clocks an image
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | without any rotation
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | (e.g. towards row 0
                                                                       | of the ndarrays)
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
        P   [xxxxxxxxx E xxxxxxxxx] [xxxxxxxxx F xxxxxxxxx] P          |
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
            [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]            |

        [] [========= 0 =========] [========= 1 =========] []
            <---------S----------   ----------S----------->


        RIGHT SIDE ROWS 4-5-6
        ---------------------

         <--------S-----------   ---------S----------->
        [] [========= 2 =========] [========= 3 =========] []          |
        /    [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  /          |
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | Direction arctic
        P   [xxxxxxxxx F xxxxxxxxx] [xxxxxxxxx E xxxxxxxxx]  P         | clocks an image
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | without any rotation
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]  |         | (e.g. towards row 0
                                                                       | of the ndarrays)
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
        P   [xxxxxxxxx G xxxxxxxxx] [xxxxxxxxx H xxxxxxxxx] P          |
        |   [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx] |          |
            [xxxxxxxxxxxxxxxxxxxxx] [xxxxxxxxxxxxxxxxxxxxx]            |

        [] [========= 0 =========] [========= 1 =========] []
            <---------S----------   ----------S----------->

        Therefore, to setup a quadrant image with the correct frame_geometry using its CCD id (from which
        we can extract its row number) and quadrant id, we need to first determine if the CCD is on the left / right
        side and then use its quadrant id ('E', 'F', 'G' or 'H') to pick the correct quadrant.
        """

        row_index = ccd_id[-1]

        if (row_index in "123") and (quadrant_id == "E"):
            return CIFrameEuclid.bottom_left(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "123") and (quadrant_id == "F"):
            return CIFrameEuclid.bottom_right(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "123") and (quadrant_id == "G"):
            return CIFrameEuclid.top_right(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "123") and (quadrant_id == "H"):
            return CIFrameEuclid.top_left(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "E"):
            return CIFrameEuclid.top_right(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "F"):
            return CIFrameEuclid.top_left(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "G"):
            return CIFrameEuclid.bottom_left(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )
        elif (row_index in "456") and (quadrant_id == "H"):
            return CIFrameEuclid.bottom_right(
                array=array,
                ci_pattern=ci_pattern,
                parallel_size=parallel_size,
                serial_size=serial_size,
                serial_prescan_size=serial_prescan_size,
                serial_overscan_size=serial_overscan_size,
                parallel_overscan_size=parallel_overscan_size,
            )

    @classmethod
    def top_left(
        cls,
        array,
        ci_pattern,
        parallel_size=2086,
        serial_size=2128,
        serial_prescan_size=51,
        serial_overscan_size=29,
        parallel_overscan_size=20,
    ):

        scans = euclid.ScansEuclid.top_left(
            parallel_size=parallel_size,
            serial_size=serial_size,
            serial_prescan_size=serial_prescan_size,
            serial_overscan_size=serial_overscan_size,
            parallel_overscan_size=parallel_overscan_size,
        )

        return CIFrame.manual(
            array=array,
            pixel_scales=0.1,
            ci_pattern=ci_pattern,
            roe_corner=(0, 0),
            scans=scans,
        )

    @classmethod
    def top_right(
        cls,
        array,
        ci_pattern,
        parallel_size=2086,
        serial_size=2128,
        serial_prescan_size=51,
        serial_overscan_size=29,
        parallel_overscan_size=20,
    ):

        scans = euclid.ScansEuclid.top_right(
            parallel_size=parallel_size,
            serial_size=serial_size,
            serial_prescan_size=serial_prescan_size,
            serial_overscan_size=serial_overscan_size,
            parallel_overscan_size=parallel_overscan_size,
        )

        return CIFrame.manual(
            array=array,
            pixel_scales=0.1,
            ci_pattern=ci_pattern,
            roe_corner=(0, 1),
            scans=scans,
        )

    @classmethod
    def bottom_left(
        cls,
        array,
        ci_pattern,
        parallel_size=2086,
        serial_size=2128,
        serial_prescan_size=51,
        serial_overscan_size=29,
        parallel_overscan_size=20,
    ):

        scans = euclid.ScansEuclid.bottom_left(
            parallel_size=parallel_size,
            serial_size=serial_size,
            serial_prescan_size=serial_prescan_size,
            serial_overscan_size=serial_overscan_size,
            parallel_overscan_size=parallel_overscan_size,
        )

        return CIFrame.manual(
            array=array,
            pixel_scales=0.1,
            ci_pattern=ci_pattern,
            roe_corner=(1, 0),
            scans=scans,
        )

    @classmethod
    def bottom_right(
        cls,
        array,
        ci_pattern,
        parallel_size=2086,
        serial_size=2128,
        serial_prescan_size=51,
        serial_overscan_size=29,
        parallel_overscan_size=20,
    ):

        scans = euclid.ScansEuclid.bottom_right(
            parallel_size=parallel_size,
            serial_size=serial_size,
            serial_prescan_size=serial_prescan_size,
            serial_overscan_size=serial_overscan_size,
            parallel_overscan_size=parallel_overscan_size,
        )

        return CIFrame.manual(
            array=array,
            pixel_scales=0.1,
            ci_pattern=ci_pattern,
            roe_corner=(1, 1),
            scans=scans,
        )



import numpy as np
from scipy.ndimage import uniform_filter
from autoarray.structures.lines.lines import Line


def find_warm_pixels(
    image,
    trail_length=9,
    n_parallel_overscan=0,
    n_serial_prescan=0,
    ignore_bad_columns=True,
    bad_column_factor=3.5,
    bad_column_loops=5,
    smooth_width=3,
    unsharp_masking_factor=4,
    flux_min=0,
    name=None,
    date=None,
):
    """ Find warm (and hot) pixels in an image.
    
    Parameters
    ----------
    image : [[float]]
        The input array of pixel values.
        
        The first dimension is the "row" index, the second is the "column" 
        index. By default (for parallel clocking), charge is transfered "up" 
        from row n to row 0 along each independent column. i.e. the readout 
        register is above row 0.
    
    trail_length : int
        The number of pixels following the warm pixel to save as a trail. The 
        same number of preceeding pixels are also included.
        
    n_parallel_overscan : int
        The number of rows in the overscan region of the input image. i.e. the 
        final rows, furthest from the readout register, beyond the physical 
        image. They should not contain warm pixels so will be ignored.
        
    n_serial_prescan : int
        The number of rows in the overscan region of the input image. i.e. the 
        first columns, closest to the readout register, before the physical 
        image. They should not contain warm pixels so will be ignored.
        
    ignore_bad_columns : bool
        Check for and ignore bad columns wiped out by extremely hot pixels.
        
    bad_column_factor : float
        Columns with a mean value more than this number of standard deviations 
        above the overall median will be discarded, to avoid columns wiped out 
        by extremely hot pixels.
    
    bad_column_loops : int
        The number of times to check for columns having means close enough to
        the overall median, updating the median each time.
    
    smooth_width : int
        The width of the window (in pixels) for calculating a smoothed image,
        used to find delta function-like warm pixels.
        
    unsharp_masking_factor : float
        Pixels must be this many times brighter than their neighbours in the 
        smoothed image to be counted as warm pixels.
        
    flux_min : float
        Pixels below this value AFTER background subtraction will be ignored.
        Set None to not ignore any pixels. Defaults to 0 to ignore pixels below 
        the background.
        
    name : str
        The name of the data used to identify it.
        
    date : float
        The Julian date for the image, for the Line objects' metadata.
    
    Returns
    -------
    warm_pixels : [Line]
        A list of the warm pixels and associated data as Line objects.
    """
    n_rows, n_columns = image.shape

    # Pixels flagged with 0 will be ignored
    where_not_ignored = np.ones_like(image)

    # List of not-ignored column indices, initially all
    good_columns = np.arange(n_columns)

    # Mean of each column
    column_means = np.mean(image, axis=0)

    # Identify bad columns to ignore, due to a really bright object or a pixel
    # hot enough to wipe everything above it, so have a high mean value
    if ignore_bad_columns:
        # Initialise to ignore all columns
        where_not_ignored *= 0

        # Remove columns with means far away from the median
        for i in range(1, bad_column_loops):
            median = np.median(column_means[good_columns])
            stddev = np.std(column_means[good_columns])
            # Keep columns with means close to the median
            good_columns = good_columns[
                abs(column_means[good_columns] - median) < bad_column_factor * stddev
            ]

        # Don't ignore the good columns
        where_not_ignored[:, good_columns] = 1

    # Subtract background
    background = 2.5 * np.median(column_means[good_columns]) - 1.5 * np.mean(
        column_means[good_columns]
    )
    image_no_bg = image - background

    # Don't ignore low-flux pixels if requested
    if flux_min is None:
        flux_min = np.nanmin(image_no_bg)

    # Unsharp mask image
    image_smooth = uniform_filter(image_no_bg, size=smooth_width)

    # Ignore the very top of the CCD since we can't get full trails
    where_not_ignored[:trail_length, :] = 0
    # Ignore parallel overscan
    where_not_ignored[-(n_parallel_overscan + trail_length) :, :] = 0
    # Ignore serial prescan
    where_not_ignored[:, :n_serial_prescan] = 0

    # Calculate the maximum of the neighbouring pixels in the same column for
    # each pixel, not including that pixel
    nearby_maxima = np.maximum.reduce(
        [
            np.roll(image_no_bg, 1, axis=0),
            np.roll(image_no_bg, 2, axis=0),
            np.roll(image_no_bg, 3, axis=0),
            np.roll(image_no_bg, 4, axis=0),
            np.roll(image_no_bg, 5, axis=0),
            np.roll(image_no_bg, 6, axis=0),
            np.roll(image_no_bg, 7, axis=0),
            np.roll(image_no_bg, 8, axis=0),
            np.roll(image_no_bg, 9, axis=0),
            np.roll(image_no_bg, -1, axis=0),
            np.roll(image_no_bg, -2, axis=0),
            np.roll(image_no_bg, -3, axis=0),
            np.roll(image_no_bg, -4, axis=0),
            np.roll(image_no_bg, -5, axis=0),
            np.roll(image_no_bg, -6, axis=0),
            np.roll(image_no_bg, -7, axis=0),
            np.roll(image_no_bg, -8, axis=0),
            np.roll(image_no_bg, -9, axis=0),
        ]
    )

    # Identify warm pixels
    warm_pixel_locations = np.argwhere(
        # Not in ignored regions
        (where_not_ignored.astype(bool))
        # Local maximum
        & (image_no_bg > nearby_maxima)
        & (image_no_bg > np.roll(image_no_bg, 1, axis=0))
        & (image_no_bg > np.roll(image_no_bg, -1, axis=0))
        # Still local maximum after unsharp masking
        & (image_no_bg > unsharp_masking_factor * np.roll(image_smooth, 1, axis=0))
        & (image_no_bg > unsharp_masking_factor * np.roll(image_smooth, -1, axis=0))
        & (image_no_bg > unsharp_masking_factor * np.roll(image_smooth, 1, axis=1))
        & (image_no_bg > unsharp_masking_factor * np.roll(image_smooth, -1, axis=1))
        # Above the minimum flux above background
        & (image_no_bg >= flux_min)
    )
    n_warm_pixels = len(warm_pixel_locations)

    if n_warm_pixels == 0:
        return []

    # Assemble the list of warm pixel data
    warm_pixels = []
    for location in warm_pixel_locations:
        row, column = location
        warm_pixels.append(
            Line(
                line=image[row - trail_length : row + trail_length, column],
                name=name,
                #                noise_map=
                location=[row, column],
                date=date,
                background=background,
            )
        )

    return warm_pixels



from autofit.tools.pipeline import ResultsCollection
from autofit.non_linear.abstract_search import Result
from autofit.mapper.model_mapper import ModelMapper

from .hyper_noise_phase import HyperNoisePhase
from .hyper_phase import HyperPhase


class CombinedHyperPhase(HyperPhase):
    def __init__(self, phase, hyper_phase_classes: (type,) = tuple()):
        """
        A hyper_combined hyper_galaxy phase that can run zero or more other hyper_galaxy phases after the initial phase is
        run.

        Parameters
        ----------
        phase : phase.PhaseCIImaging
            The phase wrapped by this hyper_galaxy phase
        hyper_phase_classes
            The classes of hyper_galaxy phases to be run following the initial phase
        """
        super().__init__(phase=phase, hyper_name="hyper_combined")
        self.hyper_phases = list(map(lambda cls: cls(phase), hyper_phase_classes))

    @property
    def phase_names(self) -> [str]:
        """
        The names of phases included in this hyper_combined phase
        """
        return [phase.hyper_name for phase in self.hyper_phases]

    def run(
        self,
        datasets,
        clocker=None,
        pool=None,
        results: ResultsCollection = None,
        **kwargs
    ) -> Result:
        """
        Run the regular phase followed by the hyper_galaxy phases. Each result of a hyper_galaxy phase is attached to the
        overall result object by the hyper_name of that phase.

        Finally, a phase in run with all of the model results from all the individual hyper_galaxy phases.

        Parameters
        ----------
        positions
        mask
        datasets
            The instrument
        results
            Results from previous phases
        kwargs

        Returns
        -------
        result
            The result of the regular phase, with hyper_galaxy results attached by associated hyper_galaxy names
        """
        results = results.copy() if results is not None else ResultsCollection()
        result = self.phase.run(datasets, results=results, clocker=clocker, **kwargs)
        results.add(self.phase.name, result)

        for phase in self.hyper_phases:
            hyper_result = phase.run_hyper(
                datasets=datasets, pool=pool, results=results, **kwargs
            )
            setattr(result, phase.hyper_name, hyper_result)

        setattr(
            result,
            self.hyper_name,
            self.run_hyper(datasets=datasets, pool=pool, results=results),
        )
        return result

    def combine_variables(self, result) -> ModelMapper:
        """
        Combine the model objects from all previous results in this hyper_combined hyper_galaxy phase.

        Iterates through the hyper_galaxy names of the included hyper_galaxy phases, extracting a result
        for each name and adding the model of that result to a new model.

        Parameters
        ----------
        result
            The last result (with attribute results associated with phases in this phase)

        Returns
        -------
        combined_variable
            A model object including all variables from results in this phase.
        """
        model = ModelMapper()
        for name in self.phase_names:
            model += getattr(result, name).model
        return model

    def run_hyper(self, datasets, pool, results, **kwargs) -> Result:
        model = self.combine_variables(result=results.last)

        phase = self.make_hyper_phase()
        phase.search.tag = ""
        phase.search.model = model

        phase.tag = ""

        return phase.run(
            datasets=datasets, results=results, pool=pool, clocker=results.last.clocker
        )



from autoconf import conf
from autofit.non_linear.paths import Paths
from autoarray.plot import mat_objs, plotters
from autocti.plot import ci_imaging_plots, ci_fit_plots


def setting(section, name):
    return conf.instance["visualize"]["plots"][section][name]


def plot_setting(section, name):
    return setting(section, name)


class AbstractVisualizer:
    def __init__(self):

        self.include = plotters.Include()

    @staticmethod
    def plotter_from_paths(paths: Paths, subfolders=None, format="png"):
        if subfolders is None:
            return plotters.Plotter(
                output=mat_objs.Output(path=f"{paths.image_path}", format=format)
            )
        return plotters.Plotter(
            output=mat_objs.Output(
                path=f"{paths.image_path}/{subfolders}", format=format
            )
        )

    @staticmethod
    def sub_plotter_from_paths(paths: Paths):

        return plotters.SubPlotter(
            output=mat_objs.Output(path=f"{paths.image_path}/subplots", format="png")
        )


class PhaseDatasetVisualizer(AbstractVisualizer):
    def __init__(self, masked_dataset):

        super().__init__()

        self.masked_dataset = masked_dataset

        self.plot_subplot_dataset = plot_setting("dataset", "subplot_dataset")
        self.plot_dataset_data = plot_setting("dataset", "data")
        self.plot_dataset_noise_map = plot_setting("dataset", "noise_map")
        self.plot_dataset_signal_to_noise_map = plot_setting(
            "dataset", "signal_to_noise_map"
        )
        self.plot_dataset_ci_pre_cti = plot_setting("dataset", "ci_pre_cti")
        self.plot_dataset_cosmic_ray_map = plot_setting("dataset", "cosmic_ray_map")

        self.plot_fit_all_at_end_png = plot_setting("fit", "all_at_end_png")
        self.plot_fit_all_at_end_fits = plot_setting("fit", "all_at_end_fits")
        self.plot_subplot_fit = plot_setting("fit", "subplot_fit")
        self.plot_subplot_residual_maps = plot_setting("fit", "subplot_residual_maps")
        self.plot_subplot_normalized_residual_maps = plot_setting(
            "fit", "subplot_normalized_residual_maps"
        )
        self.plot_subplot_chi_squared_maps = plot_setting(
            "fit", "subplot_chi_squared_maps"
        )

        self.plot_fit_data = plot_setting("fit", "data")
        self.plot_fit_noise_map = plot_setting("fit", "noise_map")
        self.plot_fit_signal_to_noise_map = plot_setting("fit", "signal_to_noise_map")
        self.plot_fit_ci_pre_cti = plot_setting("fit", "ci_pre_cti")
        self.plot_fit_ci_post_cti = plot_setting("fit", "ci_post_cti")
        self.plot_fit_residual_map = plot_setting("fit", "residual_map")
        self.plot_fit_normalized_residual_map = plot_setting(
            "fit", "normalized_residual_map"
        )
        self.plot_fit_chi_squared_map = plot_setting("fit", "chi_squared_map")


class PhaseCIImagingVisualizer(PhaseDatasetVisualizer):
    def __init__(self, masked_dataset):

        super(PhaseCIImagingVisualizer, self).__init__(masked_dataset=masked_dataset)

        self.plot_ci_fit_noise_scaling_maps_list = plot_setting(
            "fit", "noise_scaling_maps"
        )

    @property
    def masked_imaging(self):
        return self.masked_dataset

    def visualize_ci_imaging(self, paths: Paths):

        plotter = self.plotter_from_paths(paths=paths, subfolders="ci_imaging")

        ci_imaging_plots.individual(
            ci_imaging=self.masked_imaging.imaging,
            plot_image=self.plot_dataset_data,
            plot_noise_map=self.plot_dataset_noise_map,
            plot_signal_to_noise_map=self.plot_dataset_signal_to_noise_map,
            plot_ci_pre_cti=self.plot_dataset_ci_pre_cti,
            plot_cosmic_ray_map=self.plot_dataset_cosmic_ray_map,
            include=self.include,
            plotter=plotter,
        )

        sub_plotter = self.sub_plotter_from_paths(paths=paths)

        if self.plot_subplot_dataset:
            ci_imaging_plots.subplot_ci_imaging(
                ci_imaging=self.masked_imaging.imaging,
                include=self.include,
                sub_plotter=sub_plotter,
            )

    def visualize_ci_imaging_lines(self, paths: Paths, line_region):

        sub_plotter = self.sub_plotter_from_paths(paths=paths)

        if self.plot_subplot_dataset:

            sub_plotter = sub_plotter.plotter_with_new_output(
                filename=f"subplot_ci_lines_{line_region}"
            )

            ci_imaging_plots.subplot_ci_lines(
                ci_imaging=self.masked_imaging.imaging,
                line_region=line_region,
                include=self.include,
                sub_plotter=sub_plotter,
            )

        plotter = self.plotter_from_paths(
            paths=paths, subfolders=f"ci_imaging_{line_region}"
        )

        ci_imaging_plots.individual_ci_lines(
            ci_imaging=self.masked_imaging.imaging,
            line_region=line_region,
            plot_image=self.plot_dataset_data,
            plot_noise_map=self.plot_dataset_noise_map,
            plot_signal_to_noise_map=self.plot_dataset_signal_to_noise_map,
            plot_ci_pre_cti=self.plot_dataset_ci_pre_cti,
            include=self.include,
            plotter=plotter,
        )

    def visualize_ci_fit(self, paths: Paths, fit, during_analysis):

        plotter = self.plotter_from_paths(paths=paths, subfolders="fit_ci_imaging")

        ci_fit_plots.individuals(
            fit=fit,
            plot_image=self.plot_fit_data,
            plot_noise_map=self.plot_fit_noise_map,
            plot_signal_to_noise_map=self.plot_fit_signal_to_noise_map,
            plot_ci_pre_cti=self.plot_fit_ci_pre_cti,
            plot_ci_post_cti=self.plot_fit_ci_post_cti,
            plot_residual_map=self.plot_fit_residual_map,
            plot_normalized_residual_map=self.plot_fit_normalized_residual_map,
            plot_chi_squared_map=self.plot_fit_chi_squared_map,
            include=self.include,
            plotter=plotter,
        )

        if not during_analysis:

            if self.plot_fit_all_at_end_png:
                ci_fit_plots.individuals(
                    fit=fit,
                    plot_image=True,
                    plot_noise_map=True,
                    plot_signal_to_noise_map=True,
                    plot_ci_pre_cti=True,
                    plot_ci_post_cti=True,
                    plot_residual_map=True,
                    plot_normalized_residual_map=True,
                    plot_chi_squared_map=True,
                    include=self.include,
                    plotter=plotter,
                )

            if self.plot_fit_all_at_end_fits:

                self.visualize_fit_in_fits(paths=paths, fit=fit)

        sub_plotter = self.sub_plotter_from_paths(paths=paths)

        if self.plot_subplot_fit:
            ci_fit_plots.subplot_ci_fit(
                fit=fit, include=self.include, sub_plotter=sub_plotter
            )

    def visualize_ci_fit_lines(self, paths: Paths, fit, line_region, during_analysis):

        sub_plotter = self.sub_plotter_from_paths(paths=paths)

        if self.plot_subplot_fit:

            sub_plotter = sub_plotter.plotter_with_new_output(
                filename=f"subplot_ci_fit_lines_{line_region}"
            )

            ci_fit_plots.subplot_fit_lines(
                fit=fit,
                line_region=line_region,
                include=self.include,
                sub_plotter=sub_plotter,
            )

        plotter = self.plotter_from_paths(
            paths=paths, subfolders=f"fit_ci_imaging_{line_region}"
        )

        ci_fit_plots.individuals_lines(
            fit=fit,
            line_region=line_region,
            plot_image=self.plot_fit_data,
            plot_noise_map=self.plot_fit_noise_map,
            plot_signal_to_noise_map=self.plot_fit_signal_to_noise_map,
            plot_ci_pre_cti=self.plot_fit_ci_pre_cti,
            plot_ci_post_cti=self.plot_fit_ci_post_cti,
            plot_residual_map=self.plot_fit_residual_map,
            plot_normalized_residual_map=self.plot_fit_normalized_residual_map,
            plot_chi_squared_map=self.plot_fit_chi_squared_map,
            include=self.include,
            plotter=plotter,
        )

        if not during_analysis:

            if self.plot_fit_all_at_end_png:
                ci_fit_plots.individuals_lines(
                    fit=fit,
                    line_region=line_region,
                    plot_image=True,
                    plot_noise_map=True,
                    plot_signal_to_noise_map=True,
                    plot_ci_pre_cti=True,
                    plot_ci_post_cti=True,
                    plot_residual_map=True,
                    plot_normalized_residual_map=True,
                    plot_chi_squared_map=True,
                    include=self.include,
                    plotter=plotter,
                )

    def visualize_multiple_ci_fits_subplots(self, paths: Paths, fits):

        sub_plotter = self.sub_plotter_from_paths(paths=paths)

        if self.plot_subplot_residual_maps:
            ci_fit_plots.subplot_residual_maps(fits=fits, sub_plotter=sub_plotter)

        if self.plot_subplot_normalized_residual_maps:
            ci_fit_plots.subplot_normalized_residual_maps(
                fits=fits, sub_plotter=sub_plotter
            )

        if self.plot_subplot_chi_squared_maps:
            ci_fit_plots.subplot_chi_squared_maps(fits=fits, sub_plotter=sub_plotter)

    def visualize_multiple_ci_fits_subplots_lines(
        self, paths: Paths, fits, line_region
    ):

        if self.plot_subplot_residual_maps:

            sub_plotter = self.sub_plotter_from_paths(paths=paths)
            sub_plotter = sub_plotter.plotter_with_new_output(
                filename=f"subplot_residual_maps_lines_{line_region}"
            )

            ci_fit_plots.subplot_residual_map_lines(
                fits=fits, line_region=line_region, sub_plotter=sub_plotter
            )

        if self.plot_subplot_normalized_residual_maps:

            sub_plotter = self.sub_plotter_from_paths(paths=paths)
            sub_plotter = sub_plotter.plotter_with_new_output(
                filename=f"subplot_normalized_residual_maps_lines_{line_region}"
            )

            ci_fit_plots.subplot_normalized_residual_map_lines(
                fits=fits, line_region=line_region, sub_plotter=sub_plotter
            )

        if self.plot_subplot_chi_squared_maps:

            sub_plotter = self.sub_plotter_from_paths(paths=paths)
            sub_plotter = sub_plotter.plotter_with_new_output(
                filename=f"subplot_chi_squared_maps_lines_{line_region}"
            )

            ci_fit_plots.subplot_chi_squared_map_lines(
                fits=fits, line_region=line_region, sub_plotter=sub_plotter
            )

    def visualize_fit_in_fits(self, paths: Paths, fit):

        fits_plotter = self.plotter_from_paths(
            paths=paths, subfolders="fit_imaging/fits", format="fits"
        )

        ci_fit_plots.individuals(
            fit=fit,
            plot_image=True,
            plot_noise_map=True,
            plot_signal_to_noise_map=True,
            plot_ci_pre_cti=True,
            plot_ci_post_cti=True,
            plot_residual_map=True,
            plot_normalized_residual_map=True,
            plot_chi_squared_map=True,
            include=self.include,
            plotter=fits_plotter,
        )



import autocti as ac
import autofit as af
from test_autocti.integration.tests import runner

test_type = "parallel"
test_name = "x3_species__x2_image__linked_phases"
ci_data_type = "ci_uniform"
ci_data_model = "parallel_x3"
resolution = "patch"
ci_normalizations = [84700.0]


clocker = ac.Clocker(parallel_express=2)


def make_pipeline(name, folders, search=af.DynestyStatic()):

    parallel_ccd = af.PriorModel(ac.CCD)

    parallel_ccd.full_well_depth = 8.47e4
    parallel_ccd.well_notch_depth = 1e-7

    phase1 = ac.PhaseCIImaging(
        name="phase_1",
        folders=folders,
        search=search,
        parallel_traps=[af.PriorModel(ac.TrapInstantCapture)],
        parallel_ccd=parallel_ccd,
    )

    previous_total_density = phase1.result.instance.parallel_trap[0].trap_density

    trap = af.PriorModel(ac.TrapInstantCapture)
    trap.density = af.UniformPrior(lower_limit=0.0, upper_limit=previous_total_density)

    phase2 = ac.PhaseCIImaging(
        name="phase_2",
        folders=folders,
        search=search,
        parallel_traps=3 * [trap],
        parallel_ccd=phase1.result.model.parallel_ccd,
    )

    phase2 = phase2.extend_with_hyper_noise_phases()

    phase3 = ac.PhaseCIImaging(
        name="phase_3",
        folders=folders,
        parallel_traps=phase1.result.model.parallel_trap,
        parallel_ccd=phase1.result.model.parallel_ccd,
        hyper_noise_scalar_of_ci_regions=phase1.result.hyper_combined.instance.hyper_noise_scalar_of_ci_regions,
        hyper_noise_scalar_of_parallel_trails=phase1.result.hyper_combined.instance.hyper_noise_scalar_of_parallel_trails,
        search=search,
    )

    return ac.Pipeline(name, phase1, phase2, phase3)


if __name__ == "__main__":

    import sys

    runner.run(sys.modules[__name__], clocker=clocker)



import arcticpy


class TrapInstantCapture(arcticpy.TrapInstantCapture):
    """ For the old C++ style release-then-instant-capture algorithm. """

    def __init__(self, density=0.13, release_timescale=0.25):
        """ The parameters for a single trap species.
        Parameters
        ----------
        density : float
            The density of the trap species in a pixel.
        release_timescale : float
            The release timescale of the trap, in the same units as the time
            spent in each pixel or phase (Clocker sequence).
        surface : bool
            ###
        """
        super().__init__(
            density=density, release_timescale=release_timescale, surface=False
        )



from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)

from .ci_fit import CIFitImaging
from .ci_frame import CIFrame, CIFrameEuclid
from .ci_hyper import CIHyperNoiseScalar
from .ci_imaging import CIImaging
from .ci_imaging import SettingsMaskedCIImaging
from .ci_imaging import MaskedCIImaging
from .ci_imaging import SimulatorCIImaging
from .ci_mask import SettingsCIMask
from .ci_mask import CIMask
from .ci_pattern import CIPatternUniform
from .ci_pattern import CIPatternNonUniform



import arcticpy


class Clocker(object):
    def __init__(
        self,
        iterations=1,
        parallel_roe=None,
        parallel_express=0,
        parallel_charge_injection_mode=False,
        parallel_window_range=None,
        serial_roe=None,
        serial_express=0,
        serial_window_range=None,
        time_window_range=None,
    ):
        """
        The CTI Clock for arctic clocking.
        Parameters
        ----------
        parallel_sequence : [float]
            The array or single value of the time between clock shifts.
        iterations : int
            If CTI is being corrected, iterations determines the number of times clocking is run to perform the \
            correction via forward modeling. For adding CTI only one run is required and iterations is ignored.
        parallel_express : int
            The factor by which pixel-to-pixel transfers are combined for efficiency.
        parallel_charge_injection_mode : bool
            If True, clocking is performed in charge injection line mode, where each pixel is clocked and therefore \
             trailed by traps over the entire CCD (as opposed to its distance from the CCD register).
        parallel_readout_offset : int
            Introduces an offset which increases the number of transfers each pixel takes in the parallel direction.
        """

        self.iterations = iterations

        self.parallel_roe = parallel_roe
        self.parallel_express = parallel_express
        self.parallel_charge_injection_mode = parallel_charge_injection_mode
        self.parallel_window_range = parallel_window_range

        self.serial_roe = serial_roe
        self.serial_express = serial_express
        self.serial_window_range = serial_window_range

        self.time_window_range = time_window_range

    def add_cti(
        self,
        image,
        parallel_ccd=None,
        parallel_traps=None,
        serial_ccd=None,
        serial_traps=None,
    ):

        return arcticpy.add_cti(
            image=image,
            parallel_ccd=parallel_ccd,
            parallel_roe=self.parallel_roe,
            parallel_traps=parallel_traps,
            parallel_express=self.parallel_express,
            parallel_offset=image.readout_offsets[0],
            parallel_window_range=self.parallel_window_range,
            serial_ccd=serial_ccd,
            serial_roe=self.serial_roe,
            serial_traps=serial_traps,
            serial_express=self.serial_express,
            serial_offset=image.readout_offsets[1],
            serial_window_range=self.serial_window_range,
            time_window_range=self.time_window_range,
        )

    def remove_cti(
        self,
        image,
        parallel_ccd=None,
        parallel_traps=None,
        serial_ccd=None,
        serial_traps=None,
    ):
        return arcticpy.remove_cti(
            image=image,
            iterations=self.iterations,
            parallel_ccd=parallel_ccd,
            parallel_roe=self.parallel_roe,
            parallel_traps=parallel_traps,
            parallel_express=self.parallel_express,
            parallel_offset=image.readout_offsets[0],
            parallel_window_range=self.parallel_window_range,
            serial_ccd=serial_ccd,
            serial_roe=self.serial_roe,
            serial_traps=serial_traps,
            serial_express=self.serial_express,
            serial_offset=image.readout_offsets[1],
            serial_window_range=self.serial_window_range,
            time_window_range=self.time_window_range,
        )



from autocti import exc
from autofit.non_linear import abstract_search


class ConsecutivePool(object):
    """
    Replicates the interface of a multithread pool but performs computations consecutively
    """

    @staticmethod
    def map(func, ls):
        return map(func, ls)


class Analysis(abstract_search.Analysis):
    def __init__(
        self, masked_ci_datasets, clocker, settings_cti, results=None, pool=None
    ):

        super().__init__()

        self.masked_ci_datasets = masked_ci_datasets
        self.clocker = clocker
        self.settings_cti = settings_cti
        self.pool = pool or ConsecutivePool



from autocti.mask import mask as msk
from autocti.pipeline.phase.settings import SettingsPhaseCIImaging
from autocti.pipeline.phase import abstract
from autocti.pipeline.phase import extensions
from autocti.pipeline.phase.dataset.result import Result
from autofit.non_linear.paths import convert_paths
from autofit.tools.phase import Dataset
from autofit.tools.phase_property import PhaseProperty
from autofit.tools.pipeline import ResultsCollection


class PhaseDataset(abstract.AbstractPhase):

    parallel_traps = PhaseProperty("parallel_traps")
    serial_traps = PhaseProperty("serial_traps")
    parallel_ccd = PhaseProperty("parallel_ccd")
    serial_ccd = PhaseProperty("serial_ccd")

    Result = Result

    def __init__(
        self,
        search,
        parallel_traps=None,
        parallel_ccd=None,
        serial_traps=None,
        serial_ccd=None,
        settings=SettingsPhaseCIImaging(),
    ):
        """

        A phase in an lens pipeline. Uses the set non_linear search to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        search: class
            The class of a non_linear search
        """

        super().__init__(search=search)

        self.parallel_traps = parallel_traps or []
        self.parallel_ccd = parallel_ccd
        self.serial_traps = serial_traps or []
        self.serial_ccd = serial_ccd
        self.settings = settings

    def run(self, datasets: Dataset, clocker, results=None, info=None, pool=None):
        """
        Run this phase.

        Parameters
        ----------
        positions
        mask: Mask2D
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        dataset: scaled_array.ScaledSquarePixelArray
            An masked_imaging that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """
        #    self.save_metadata(dataset=datasets)
        #    self.save_dataset(dataset=datasets)
        #    self.save_mask(masks)

        self.model = self.model.populate(results)

        results = results or ResultsCollection()

        self.modify_search_paths()

        analysis = self.make_analysis(
            datasets=datasets, clocker=clocker, results=results, pool=pool
        )

        #    attributes = self.make_attributes(analysis=analysis)
        #    self.save_attributes(attributes=attributes)

        result = self.run_analysis(analysis=analysis, info=info)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, results=None, pool=None):
        """
        Returns an lens object. Also calls the prior passing and masked_imaging modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        positions
        mask: Mask2D
            The default masks passed in by the pipeline
        dataset: im.Imaging
            An masked_imaging that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the `NonLinearSearch` calls to determine the fit of a set of values
        """
        raise NotImplementedError()

    def mask_for_analysis_from_cosmic_ray_map(self, cosmic_ray_map, mask):

        cosmic_ray_mask = (
            msk.Mask2D.from_cosmic_ray_map_buffed(
                cosmic_ray_map=cosmic_ray_map, settings=self.settings.settings_mask
            )
            if cosmic_ray_map is not None
            else None
        )

        if cosmic_ray_map is not None:
            return mask + cosmic_ray_mask

        return mask

    def modify_search_paths(self):

        self.search.paths.tag = self.settings.phase_tag

    def extend_with_hyper_noise_phases(self):
        return extensions.CombinedHyperPhase(
            phase=self, hyper_phase_classes=[extensions.HyperNoisePhase]
        )



from autofit import exc
from autoarray.exc import ArrayException, MaskException, FrameException, RegionException


class DatasetException(Exception):
    pass


class CIPatternException(Exception):
    pass


class PlottingException(Exception):
    pass


class FittingException(Exception):
    pass


class PriorException(exc.FitException):
    pass
